%{					/* -*-c-*- */
 /**********************************************************************
 @Lex-file{
    author              = "Nelson H. F. Beebe",
    version             = "0.03",
    date                = "29 October 2020",
    time                = "16:48:47 MDT",
    filename            = "texpty.l",
    address             = "Department of Mathematics
			   University of Utah
			   Salt Lake City, UT 84112
			   USA",
    telephone           = "+1 801 581 5254",
    FAX                 = "+1 801 581 4148",
    checksum            = "30122 4939 14143 122161",
    email               = "beebe@math.utah.edu (Internet)",
    email               = "beebe@math.utah.edu, beebe@acm.org,
                           beebe@computer.org (Internet)",
    codetable           = "ISO/ASCII",
    keywords            = "AmSTeX; AmSLaTeX; LaMSTeX; LaTeX; SliTeX;
			   TeX; prettyprinter",
    license             = "public domain",
    supported           = "yes",
    docstring           = "This is a simple TeX (AmSLaTeX, AmSTeX,
			   ETeX, LAmSTeX, LaTeX, plain TeX, SliTeX,
			   TeXinfo, ...) file prettyprinter.

			   Here is a brief summary; for full
			   documentation, consult the accompanying UNIX
			   manual pages:

			   Usage:
			       texpretty [-?] [-@] [-a] [-c] [-d] [-e mode]
					  [-f filename] [-h] [-i nnn]
					  [-l logfile] [-m nnn] [-n]
					  [-o filename] [-q] [-s filename] [-t]
					  [-v] [-w width]
					  [file(s)]
			   or
			       texpretty [-?] [--texinfo] [--author]
					  [--copyright] [--displaymath-all]
					  [--emacs-mode mode]
					  [--filename filename] [--help]
					  [--indent nnn] [--logfile logfile]
					  [--math-conversions nnn]
					  [--no-comment-banner]
					  [--outfile filename] [--quick]
					  [--stylefile filename]
					  [--tabular-as-verbatim] [--version]
					  [--width width]
					  [file(s)]

			   The options recognized are:

				-? (show help message)
				-@ (set TeXinfo mode)
				-a (show author)
				-c (show copyright)
				-d (output inline math like display math)
				-e (set emacs modename)
				-f (set output filename in comment banner)
				-h (show help message)
				-i (set indentation size)
				-l (set error log filename)
				-m (math mode conversions)
				-n (no comment banner)
				-o (set output filename)
				-q (no default style files)
				-s (set style filename)
				-t (copy tabular environments verbatim)
				-v (show version)
				-w (set linewidth)

			   Letter case in option names is NOT
			   significant.  On IBM PC DOS and DEC VMS, the
			   leading hyphen in option names may be
			   replaced by a slash; both hyphen and slash
			   are accepted on that system.

			   Options may be intermixed with filenames,
			   in which case, they apply only to following
			   filenames.  However, -q is always handled
			   immediately.

			   As with any simple-minded prettyprinter,
			   especially ones for natural language text,
			   which rarely conforms to a strict grammar,
			   you should check the output of the
			   prettyprinter before replacing the input by
			   it.  In particular, if your input file
			   follows the standards imposed by this
			   prettyprinter, a file difference listing may
			   be reassuring.

			   A fundamental assumption of any prettyprinter
			   is that whitespace (blank, horizontal tab,
			   vertical tab, carriage return, formfeed,
			   newline) is NOT significant.  While this is
			   generally the case, it is not so in those
			   rare TeX files where whitespace is made
			   active by \catcode modification.  Avoid using
			   texpretty on such files, or on files where
			   the comment character is something other than
			   the normal percent character.  There is no
			   problem with LaTeX verbatim macros and
			   environments, because texpretty recognizes
			   them and preserves their contents exactly.

			   Because of their low-level markup, plain TeX
			   and AmSTeX files offer fewer opportunities
			   for useful prettyprinting than do LaTeX,
			   LaMSTeX, and SliTeX.  Nevertheless, the
			   heuristic error checking provided by
			   texpretty may still be useful for catching
			   brace and dollar balance errors.

			   texpretty may be less useful for files
			   consisting only of macro definitions (e.g.
			   plain.tex, or LaTeX style files), because
			   they may already have stylized formatting,
			   because macro definitions might include
			   unclosed environments, and because there is
			   usually little structure to be exposed by
			   prettyprinting.

			   The checksum field above contains a CRC-16
			   checksum as the first value, followed by the
			   equivalent of the standard UNIX wc (word
			   count) utility output of lines, words, and
			   characters.  This is produced by Robert
			   Solovay's checksum utility.",
 }
 **********************************************************************/

	/* texpty -- prettyprint TeX files */

#if defined(DEBUG_MALLOC)
#define free	debug_free
#define malloc	debug_malloc
#define realloc	debug_realloc
#endif

#define VERSION		"0.03"		/* these MUST match version */
#define DATE		"[29-Oct-2020]"	/* and date above */
#define PROGRAM		"texpretty"

#if !defined(__DATE__)
#define __DATE__ ""
#endif /* __DATE__ */

#if !defined(__TIME__)
#define __TIME__ ""
#endif /* __TIME__ */

#if !defined(USER)
#define USER	""
#endif

#if !defined(HOSTNAME)
#define HOSTNAME ""
#endif

#include <config.h>

#if HAVE_STDIO_H
#include <stdio.h>
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_STRING_H
#include <string.h>			/* for strlen() declaration */
#endif

#if HAVE_CTYPE_H
#include <ctype.h>
#endif

#if HAVE_LIMITS_H
#include <limits.h>
#endif

#ifndef UCHAR_MAX
#define UCHAR_MAX 255
#endif

#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#if HAVE_PWD_H
#include <pwd.h>
#endif

#if defined(__NeXT__)
#undef HAVE_UNAME
#if defined(__GNUG__)

#if HAVE_LIBC_H
#include <libc.h>
#endif

#else /* NOT defined(__GNUG__) */

/***********************************************************************
This ugliness of explicit declarations arises because the installation
of gcc 2.6.3 (and earlier) on the NeXT system searches vendor-provided
header files before gcc ones; the result is function prototype
conflicts.  They can be avoided if the code is compiled with an explicit
include-file path, such as
	-I/usr/local/lib/gcc-lib/m68k-next-nextstep3/2.6.3/include/bsd
but that is an ugly solution too, so instead we just provide copies of
the declarations from the gcc libc.h that we actually need.
***********************************************************************/
extern char	*getlogin(void);
extern uid_t    getuid(void);
extern int      gethostname(char *, int);
#endif  /* NOT defined(__GNUG__) */

#else
#define HAVE_UNAME
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_SYS_UTSNAME_H
#include <sys/utsname.h>
#endif

#if HAVE_TIME_H
#include <time.h>
#endif

#define COMMENT_PREFIX	"%%% "
#define COMMENT_SUFFIX	""

#define FPRINTF		(void)fprintf
#define SPRINTF		(void)sprintf
#define PUTSTDERR(msg)	(void)fputs((msg), stderr)
#define PUTSTDOUT(msg)	(void)fputs((msg), stdout)

#if !defined(INDENT)
#define INDENT 4			/* can override at compile time */
#endif

#undef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))

#undef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))

#if !defined(MAXDELIMITERLEVEL)
#define MAXDELIMITERLEVEL 25		/* can override at compile time */
#endif

#if !defined(MAXBUF)
#define MAXBUF 2048			/* can override at compile time */
#endif

#if MAXBUF < 1024
#undef MAXBUF
#define MAXBUF 1024			/* guaranteed minimum: see man pages */
#endif

#if !defined(MAXENVIRONMENTLEVEL)
#define MAXENVIRONMENTLEVEL 25		/* can override at compile time */
#endif

#if !defined(MAXWIDTH)
#define MAXWIDTH 72			/* can override at compile time */
#endif

#define NOOP	/* NO-OP */

#if defined(MSDOS)|| defined(_MSDOS)|| defined(__MSDOS__)|| defined(__TURBOC__)
#define IBM_PC
#endif

#if defined(IBM_PC)
#define OPTION_PREFIX(c)	( ((c) == '-') || ((c) == '/') )
/* On the IBM PC, we process files in binary mode, so that Ctl-Z does
not act as an end-of-file mark. */
#define RMODE "rb"
#define WMODE "wb"

#if HAVE_FCNTL_H
#include <fcntl.h>			/* for O_BINARY in setmode() call */
#endif

#if HAVE_IO_H
#include <io.h>				/* for setmode() prototypes */
#endif


#if defined(_O_BINARY) && !defined(O_BINARY)
/* Microsoft C version 7.0 requires leading underscore on these symbols
in Standard C mode (command-line -Za switch).  Versions 5.0, 5.1, and
6.0 do not use the leading underscore.  */
#define O_BINARY _O_BINARY
#define setmode _setmode
#endif

#else  /* not IBM_PC */
#define OPTION_PREFIX(c)	((c) == '-')
#define RMODE "r"
#define WMODE "w"
#endif /* IBM_PC */

#if defined(VMS)
#undef OPTION_PREFIX
#define OPTION_PREFIX(c)	( ((c) == '-') || ((c) == '/') )
#endif

#include "stdc.h"			/* NB: must be included AFTER */
					/* all system header files, to */
					/* avoid inconsistent definitions */
					/* of const */

#if defined(IBM_PC)
#define DEFAULT_STYLE_FILE		"texpty.ini"
#define DIRECTORY_SEPARATOR		"\\"
#define HOME_DIRECTORY			"c:"
#endif

#if defined(VMS)
#define DEFAULT_STYLE_FILE		"texpretty.ini"
#define DIRECTORY_SEPARATOR		""
#define HOME_DIRECTORY			"sys$login:"
#endif

#if !defined(DEFAULT_STYLE_FILE)
#define DEFAULT_STYLE_FILE		".texprettyrc"
#endif

#if !defined(DIRECTORY_SEPARATOR)
#define DIRECTORY_SEPARATOR		"/"
#endif

#if !defined(HOME_DIRECTORY)
#define HOME_DIRECTORY			"HOME"
#endif

#if defined(__TURBOC__)

/* This code will compile fine under Turbo C 2.0, which always defines
fileno().  However, Turbo C 3.0 defines fileno() only in a non-Standard
C environment (-A command-line flag omitted).  Compilation fails under
both 2.0 and 3.0 if -A is omitted, because some header files get
included with const modifiers (flex checks for Turbo C and ensures that
const is used), and they later get included again with const defined to
an empty string (by stdc.h), resulting in function prototype declaration
conflicts.  Thus, we need to compile in the Standard C environment, and
then supply a definition of fileno() if one hasn't been provided. */

#ifndef fileno
#define fileno(f)       ((f)->fd)
#endif
#endif /* defined(__TURBOC__) */

#if defined(__cplusplus) || defined(c_plusplus)
extern "C" {
#endif

#if !defined(HAVE_FILENO_DECL)
int fileno ARGS((FILE *));
#endif

#if defined(__cplusplus) || defined(c_plusplus)
};
#endif


/* Silicon Graphics IRIX 5.3 lex supplies a 200-character input
buffer, which is much too small, so increase it. */

#if defined(YYLMAX) && ((YYLMAX + 0) < 8192)
#undef YYLMAX
#define YYLMAX 8192
#endif

typedef void (*PostOutputAction)ARGS((const char *name_));

typedef void (*StyleFunction)ARGS((void));

typedef struct
{
    const char *style;
    const char *name;
    StyleFunction action;
} Style;

typedef struct
{
    size_t max_size;
    size_t size;
    Style *table;
} StyleTable;

#if STDC
typedef void* VOIDP;
#else
typedef char* VOIDP;
#endif

typedef unsigned int	HASH_INT;	/* MUST be unsigned for hash() */
					/* to protect against overflow */
typedef struct Hash_Entry
{
    const char		*hash_key;	/* pointer to hash key string */
    VOIDP		hash_data;	/* pointer to arbitrary data */
} HASH_ENTRY;

typedef struct Hash_Table
{
    HASH_INT	hash_index;		/* index of next entry to be */
					/* examined by hash_next() */
    HASH_INT	hash_size;		/* size of hash table */
    HASH_ENTRY	*hash_table;		/* pointer to hash table */
} HASH_TABLE;

static int		brace_level = 0;
static int		brace_line[MAXDELIMITERLEVEL];
static int		bracket_level = 0;
static int		bracket_line[MAXDELIMITERLEVEL];
static const char	*default_begin_math = (const char*)NULL;
static const char	*default_begin_display_math = (const char*)NULL;
static int		display_math_mode = 0;
static int		do_banner = 1;
static const char	*default_end_math = (const char*)NULL;
static const char	*default_end_display_math = (const char*)NULL;
static const char	*emacs_mode = "TeX";
static int		environment_level = 0;
static int		environment_line[MAXENVIRONMENTLEVEL];
static char		*environment_name[MAXENVIRONMENTLEVEL];
static const char	*filename = "";
static int		g_argc;
static char		**g_argv;
static int		g_argk;
static int		g_errors;
static int		in_comment = 0;
static int		in_tabular = 0;
static int		in_texinfo = 0;
int			in_verbatim = 0;
static int		indent = INDENT;
static char		line_buffer[MAXBUF];
static int		line_number = 1;
static int		math_conversions = 0;
static int		math_level = 0;
static int		math_line[MAXENVIRONMENTLEVEL];
static int		max_width = MAXWIDTH;
static int		next_position = 0; /* index into line_buffer[] */
static int		parenthesis_level = 0;
static int		parenthesis_line[MAXDELIMITERLEVEL];
static PostOutputAction post_output_action;

#if defined(DEBUG)
static char		prefix[4];	/* for DEBUGGING */
#endif

static HASH_TABLE	*style_hash_table = (HASH_TABLE *)NULL;
static StyleTable	style_table = { 0, 0, (Style*)NULL };
static int		tabular_like_verbatim = 0;
static int		wrap_before_chars[UCHAR_MAX];


static void		add_style ARGS((const char *style, const char *name));
static void		adjust_brace_level ARGS((int n));
static void		adjust_bracket_level ARGS((int n));
void			adjust_environment_level ARGS((int n,
						       const char *command));
static void		adjust_list_level ARGS((int n, const char *command));
static void		adjust_math_level ARGS((int n));
static void		adjust_parenthesis_level ARGS((int n));
static void		ampersand ARGS((void));
static void		author ARGS((void));
static void		bad_option ARGS((const char *option));
static void		begin_brace ARGS((void));
static void		begin_bracket ARGS((void));
static void		begin_display_math ARGS((void));
static void		begin_environment ARGS((void));
static void		begin_list ARGS((void));
static void		begin_math ARGS((void));
static void		begin_parenthesis ARGS((void));
static void		begin_tabular ARGS((void));
static void		begin_verbatim ARGS((void));
static void		begin_verbatim_output ARGS((void));
static void		blank ARGS((void));
static void		check_brace_level_is_zero ARGS((const char *msg,
							int reset));
static void		check_bracket_level_is_zero ARGS((const char *msg,
							int reset));
static void		check_delimiter_level ARGS((int c));
static void		check_environment_level_is_zero ARGS((const char *msg,
							int reset));
static void		check_environment_name_match ARGS((const char *begin,
							   const char *end));
static void		check_math_level_is_zero ARGS((const char *msg,
							int reset));
static void		check_parenthesis_level_is_zero ARGS((const char *msg,
							int reset));
static void		cite ARGS((void));
static void		command_argument ARGS((void));
static void		command_line ARGS((void));
static void		comment_and_newline_before ARGS((void));
static void		copyright ARGS((void));
static void		copy_argument ARGS((void));
extern void		copy_tabular ARGS((void));
static void		copy_to_end_of_line ARGS((void));
static void		copy_verbatim ARGS((void));

#if defined(DEBUG_MALLOC)
static void		debug_free ARGS((VOIDP ptr));
static VOIDP		debug_malloc ARGS((size_t size));
static VOIDP		debug_realloc ARGS((VOIDP ptr, size_t size));
#endif

static void		delete_indentation ARGS((void));
void			delete_spaces ARGS((void));
static void		display_math ARGS((void));
static void		do_command_line_argument ARGS((void));
static void		do_initialization_files ARGS((void));
static void		do_macro_arguments ARGS((const char *pattern));
static void		do_nothing ARGS((const char *pattern));
static void		do_style_file ARGS((const char *the_filename,
					    int warn_if_no_file));
static void		do_style_line ARGS((const char *source,
					    int style_line_number,
					    char *line));
static char		*dupstr ARGS((const char *s));
static void		end_brace ARGS((void));
static void		end_bracket ARGS((void));
static void		end_display_math ARGS((void));
static void		end_environment ARGS((void));
static void		end_list ARGS((void));
static void		end_math ARGS((void));
static void		end_parenthesis ARGS((void));
static void		end_tabular ARGS((void));
static void		end_verbatim ARGS((const char *token));
static void		end_verbatim_output ARGS((void));
static void		environment ARGS((void));
static void		error ARGS((const char *msg));
static void		free_hash_table ARGS((void));
static void		free_style_table ARGS((void));
static char 		*get_macro ARGS((const char *pattern));
VOIDP			get_memory ARGS((VOIDP ptr, size_t size));
static char 		*get_name ARGS((const char *command));
static Style		*get_style_by_name ARGS((const char *name));
static Style		*get_style_by_style ARGS((const char *name));
static char		*get_token ARGS((char *name, const char *set,
					 char *overwritten_char));
static HASH_INT		hash ARGS((const char *key, const HASH_TABLE *table));
static HASH_TABLE	*hash_alloc ARGS((HASH_INT size));
static void		hash_free ARGS((HASH_TABLE *table));
static HASH_ENTRY	*hash_install ARGS((const char* key, VOIDP data,
					    HASH_TABLE *table));
static HASH_ENTRY	*hash_lookup ARGS((const char* key, HASH_TABLE *table));
static int		indentation_size ARGS((void));
static void		index_and_glossary ARGS((void));
static void		initialize ARGS((void));

#if defined(__cplusplus) || defined(c_plusplus)
			/* Sun Solaris 2.x lex renames input() to yyinput() */
#define Input()		yyinput()
#else
#define Input()		input()
#endif /* defined(__cplusplus) */

int			input_char ARGS((void));

#define Is_Control_Sequence(s)	(Is_TeX(s) || Is_TeXinfo(s))
#define Is_Control_Sequence_Character(c) (isalpha(c) || ((c) == '@'))
#define Is_NonEmpty_Line()	(line_length() > indentation_size())
#define Is_TeX(s)		(s[0] == '\\')
#define Is_TeXinfo(s)		(s[0] == '@')

int			last_char ARGS((int offset));

#if defined(FLEX_SCANNER)
#define LEXEOF(c)	((c) == EOF)	/* flex end-of-file test */
#else
#define LEXEOF(c)	((c) == 0)	/* lex end-of-file test */
#endif

static void		line_end ARGS((void));
static int		line_length ARGS((void));
static void		list ARGS((void));
static void		list_item ARGS((void));
static const char	*long_to_short_option ARGS((const char *option));
static void		make_hash_table ARGS((void));
static void		math ARGS((void));
static void		newline_after ARGS((void));
static void		newline_before ARGS((void));
static long		next_prime ARGS((long n));
static void		normalize_space ARGS((char *s));
static void		obey ARGS((void));
static void		out_banner ARGS((void));
static void		out_blank ARGS((void));
static void		out_buffer ARGS((void));
void			out_char ARGS((int c));
static void		out_comment ARGS((void));
static void		out_control_sequence ARGS((void));
static void		out_end_math_punctuation ARGS((const char *prefix));
void			out_indentation ARGS((void));
static void		out_leading_blanks ARGS((int nblanks));
void			out_linebreak ARGS((void));
static void		out_linenumbers ARGS((const int *array,
					      const char **name, int n));
static void		out_newline ARGS((void));
void			out_string ARGS((const char *s));
static void		out_token_on_new_line ARGS((const char *token));
static void		out_token_on_own_line ARGS((const char *token));

#define out_verbatim_char(c) out_char(c)

static void		out_verbatim_string ARGS((const char *s));
static void		out_yytext ARGS((void));
static void		remove_separating_blanks ARGS((void));
static int		reopen ARGS((void));
static void		sectional_division ARGS((void));
static void		sectional_division_line ARGS((void));
static void		standalone ARGS((void));
static void		tabular ARGS((void));
static void		trim_line ARGS((int n));
static void		unput_char ARGS((int c));
static void		usage ARGS((void));
static void		verb ARGS((void));
static void		verbatim ARGS((void));
static void		version ARGS((void));
static void		warning ARGS((const char *msg));
static void		wrapup ARGS((void));

#if defined(yywrap)
#undef yywrap		/* this is a macro in some older flex versions */
#endif

#ifdef __cplusplus
extern "C"
#endif
int			yywrap ARGS((void));

Style default_style[] =
{
    { "chapter",	(const char*)NULL,	sectional_division },
    { "chapter-line",	(const char*)NULL,	sectional_division_line },
    { "cite",		(const char*)NULL,	cite },
    { "command-argument", (const char*)NULL,	command_argument },
    { "command-line",   (const char*)NULL,	command_line },
    { "default",	(const char*)NULL,	out_yytext },
    { "comment",	(const char*)NULL,	out_comment },
    { "displaymath",	(const char*)NULL,	display_math },
    { "environment",	(const char*)NULL,	environment },
    { "footnote",	(const char*)NULL,	comment_and_newline_before },
    { "index",		(const char*)NULL,	index_and_glossary },
    { "list",		(const char*)NULL,	list },
    { "list-item",	(const char*)NULL,	list_item },
    { "math",		(const char*)NULL,	math },
    { "newline-after",	(const char*)NULL,	newline_after },
    { "newline-before",	(const char*)NULL,	newline_before },
    { "standalone",	(const char*)NULL,	standalone },
    { "tabular",	(const char*)NULL,	tabular },
    { "verb",		(const char*)NULL,	verb},
    { "verbatim",	(const char*)NULL,	verbatim},
    { (const char *)NULL,(const char*)NULL,	(StyleFunction)NULL }
};

/***********************************************************************
Explicit specification of large numbers of macro names in the lex
patterns led to problems: lex could no longer process the file, though
flex still could, and the size of the generated yy_xxx[] lexer tables
became too large for compilation on the IBM PC 64KB-segmented memory
space.

Movement of macro names into the builtin_style[] table shortens the
length of this file by about 400 lines, at the (modest) expense of
run-time parsing of this table via main() -> initialize() ->
do_style_line(), and the introduction of hash table support for fast
lookup.  It also paves the way for possible implementation of a run-time
choice of TeX dialect, instead of the current recognize-all-dialects
approach.
***********************************************************************/

const char *builtin_style[] =
{
    /* NB: These strings may be no longer than MAXBUF (>= 1024), */
    /* because initialize() must be able to copy them into a fixed */
    /* buffer, line[MAXBUF+2].   They may also not exceed 512 */
    /* characters in length, in order to satisfy a Microsoft C 5.x */
    /* limit, sigh... */
    /* Macros must consist entirely of letters, and require no */
    /* special argument handling.  Such cases must be dealt with */
    /* by explicit lex patterns below.   For readability, keep */
    /* them ordered by package, class, and name. */


    /* AmSLaTeX (extension of LaTeX) */
    "newline-before : \\DeclareMathOperator \\DeclareMathSymbol \
    \\DeclareSymbolFont \\numberwithin \\swapnumbers \\newtheoremstyle \
    \\theoremstyle",


    /* AmSLaTeX */
    "standalone : \\allowdisplaybreaks",


    /* AmSTeX */
    "environment : \\CD \\Refs \\Sb \\Sp \\Vmatrix \\abstract \\affil \
    \\align \\alignat \\aligned \\alignedat \\block \\bmatrix \\cases \
    \\cfrac \\comment \\curraddr \\dedicatory \\definition \\demo \
    \\document \\email \\example \\gather \\gathered \\graf \\head \
    \\keywords \\matrix \\multline \\pmatrix \\proclaim \\remark \
    \\smallmatrix \\specialhead \\split \\subhead \\subjclass \
    \\subsubhead \\toc \\topmatter \\translator \\vmatrix \\xalignat",

    "environment : \\xxalignat \\endCD \\endRefs \\endSb \\endSp \
    \\endVmatrix \\endabstract \\endaffil \\endalign \\endalignat \
    \\endaligned \\endalignedat \\endblock \\endbmatrix \\endcases",

    "environment : \\endcfrac \\endcomment \\endcurraddr \
    \\enddedicatory \\enddefinition \\enddemo \\enddocument \\endemail \
    \\endexample \\endgather \\endgathered \\endgraf \\endhead \
    \\endkeywords \\endmatrix \\endmultline \\endpmatrix \\endproclaim \
    \\endremark \\endsmallmatrix \\endspecialhead \\endsplit \
    \\endsubhead \\endsubjclass \\endsubsubhead \\endtoc \
    \\endtopmatter \\endtranslator \\endvmatrix \\endxalignat \
    \\endxxalignat",


    /* AmSTeX */
    "footnote : \\plainfootnote \\plainproclaim",


    /* AmSTeX */
    "list : \\roster \\endroster",


    /* AmSTeX */
    "list-item : \\runinitem",


    /* AmSTeX conflicts with LaTeX: LaTeX takes precedence, and */
    /* AmSTeX loses some leading indentation */
    "newline-before : \\endaddress \\endauthor \\endcaption \\enddate \
    \\endref \\endthanks \\endtitle",


    /* AmSTeX */
    "newline-before : \\adjustfootnotemark \\book \\bookinfo \
    \\bookinquotes \\botcaption \\by \\bysame \\captionwidth \\define \
    \\ed \\eds \\finalinfo \\inbook \\issue \\jour \\lang \\moreref \
    \\newsymbol \\no \\noquotes \\page \\pages \\paper \\paperinfo \
    \\paperinquotes \\parshape \\preaffil \\preauthor \\predate \
    \\predefine \\prepaper \\pretitle \\publ \\publaddr \\redefine \
    \\shoveleft \\shoveright \\toappear \\topcaption \\transl \
    \\undefine \\vol \\yr",


    /* AmSTeX */
    "standalone : \\BlackBoxes \\CenteredTagsOnSplits \\LimitsOnInts \
    \\LimitsOnNames \\LimitsOnSums \\Monograph \\NoBlackBoxes \
    \\NoPageNumbers \\NoRunningHeads \\Runinitem \\TagsAsMath \
    \\TagsAsText \\TagsOnLeft \\TagsOnRight \\TopOrBottomTagsOnSplits",

    "standalone : \\UseAMSsymbols \\UseBibTeX \\endinsert \\foldedpar \
    \\galleys \\ininbook \\loadbold \\loadeufb \\loadeufm \\loadeurb \
    \\loadeurm \\loadeusb \\loadeusm \\loadmsam \\loadmsbm \\midinsert \
    \\pageinsert \\printoptions \\showallocations \\syntax \\topinsert",


    /* Extended plain TeX (ETeX) */
    "footnote : \\numberedfootnote",


    /* ETeX */
    "index : \\idx \\idxmarked \\idxname \\idxsubmarked \\sidx \
    \\sidxmarked \\sidxname \\sidxsubmarked",


    /* ETeX */
    "list : \\numberedlist \\endnumberedlist \\orderedlist \
    \\endorderedlist \\unorderedlist \\endunorderedlist",


    /* ETeX */
    "newline-before : \\defineindex \\definecontentsfile \\center \
    \\columnfill \\doublecolumns \\edefappend \\flushleft \\flushright \
    \\for \\iffileexists \\innerdef \\innerinnerdef \\innernewbox \
    \\innernewcount \\innernewdimen \\innernewfam \\innernewhelp",

    "newline-before : \\innernewif \\innernewinsert \\innernewmuskip \
    \\innernewread \\innernewskip \\innernewtoks \\innernewwrite \
    \\listing \\makecolumns \\quadcolumns \\readindexfile \
    \\readtocfile \\testfileexistence \\tocchapterentry \
    \\tocsectionentry \\triplecolumns \\writenumberedtocentry \
    \\writetocentry",


    /* ETeX */
    "verb : \\verbatim",


    /* LAmSTeX */
    "chapter : \\docstyle \\subtopic \\topic",


    /* LAmSTeX */
    "environment : \\Figure \\Figurepair \\Figuretriple \\HL \\Table \
    \\bdmatrix \\claim \\heading \\island \\makebib \\partition \
    \\endFigure \\endFigurepair \\endFiguretriple \\endHL \\endTable \
    \\endbdmatrix \\endclaim \\endheading \\endisland \\endmakebib \
    \\endpartition",


    /* LAmSTeX */
    "list : \\bib \\bullist \\describe \\list \\margins \\endbib \
    \\endbullist \\enddescribe \\endlist \\endmargins",


    /* LAmSTeX */
    "newline-before : \\Cgaps \\Entry \\Entryxref \\LETTER \\Morexref \
    \\Noexpand \\Nonexpanding \\Page \\PageSpan \\Pagespan \
    \\PostCDSpace \\PreCDSpace \\Rgaps \\Topage \\Xref \\cgaps \
    \\cleartable \\counter \\everytable \\ex \\exs \\flushpar \\fnote \
    \\foottext \\hL \\hdashed \\hl \\hls \\htablelines \\iabbrev \
    \\idefine \\litbackslash \\litdelimiter \\mainfile \\makepiece \
    \\manyby \\measuretable \\modifyfootnote \\nameHL \\namehl",

    "newline-before : \\newclaim \\newisland \\note \\pageorder \
    \\postCDspace \\postdocstyle \\preCDspace \\predocstyle \\pullin \
    \\pullinmore \\purge \\readaux \\rgaps \\runningchapter \
    \\runningsection \\shortenclaim \\showstored \\sss \\storetable \
    \\tablewidth \\tbldocstyle \\tdefine \\toclevel \\tredefine \\tss \
    \\unpurge \\usetable \\vleft \\vright \\vs \\vsolid \\vtablelines",


    /* LAmSTeX */
    "standalone : \\Figureproofing \\FlushedFigs \\Initialize \\NS \
    \\NoFlushedFigs \\RefWarnings \\alldq \\boxedtables \\columnbreak \
    \\continuelist \\figureproofing \\indexfile \\indexproofing \
    \\inlevel \\keepitem \\makelistFigures \\makelistTables \
    \\makelistfigures \\makelisttables \\maketoc \\newcolumn",

    "standalone : \\noFigureproofing \\nocolumnbreak \
    \\nofigureproofing \\noshowsecondpass \\opentables \\outlevel \
    \\shortlastcolumn \\showcolwidths \\showsecondpass \\sides \
    \\ssizeCDlabels \\tocfile \\tsizeCDlabels",


    /* LAmSTeX */
    "verb : \\Lit \\lit",


    /* LaTeX and LAmSTeX */
    "chapter : \\appendix \\backmatter \\chapter \\documentclass \
    \\documentstyle \\frontmatter \\mainmatter \\paragraph \\part \
    \\section \\subparagraph \\subsection \\subsubsection",


    /* LaTeX and LAmSTeX */
    "cite : \\cite \\nocite",


    /* LaTeX */
    "footnote : \\footnotetext \\label",


    /* LaTeX */
    "index : \\glossary \\index",


    /* LaTeX */
    "newline-after : \\kill",


    /* LaTeX and LAmSTeX */
    "newline-before : \\author \\date \\thanks \\title",


    /* LaTeX */
    "newline-before : \\address \\caption \\closing \\glossaryentry \
    \\include \\includeonly \\indexentry \\makeglossary \\makeindex \
    \\marginpar \\markboth \\markright \\multiput \\newblock \
    \\newboolean \\newcommand \\newcounter \\newenvironment \
    \\newlength \\newsavebox \\newtheorem \\opening \\printindex \
    \\providecommand \\put \\renewcommand \\renewenvironment \
    \\signature \\typein \\typeout \\usepackage \\vspace",


    /* LaTeX */
    "standalone : \\bigpagebreak \\cleardoublepage \\clearpage \
    \\flushbottom \\fussy \\hline \\indexspace \\listfiles \
    \\listoffigures \\listoftables \\maketitle \\medpagebreak \
    \\newline \\newpage \\nofiles \\normalbottom \\normalmarginpar \
    \\onecolumn \\onlynotes \\onlyslides \\raggedleft \
    \\reversemarginpar \\sloppy \\smallpagebreak \\tableofcontents \
    \\twocolumn",


    /* LaTeX */
    "verb: \\path \\verb",


    /* plain TeX */
    "chapter : \\beginchapter \\beginsection \\endchapter",


    /* plain TeX */
    "environment : \\bgroup \\begingroup \\egroup \\endgroup",


    /* plain TeX */
    "footnote : \\or",


    /* plain TeX */
    "footnote : \\footnote \\vfootnote",


    /* plain TeX */
    "newline-after : \\cr \\crcr \\endline",


    /* plain TeX */
    "newline-before : \\chardef \\closein \\closeout \\countdef \\def \
    \\dimendef \\edef \\else \\endinput \\eqalign \\eqalignno \
    \\errmessage \\fi \\futurelet \\futurenonspacelet \\gdef \\global \
    \\halign \\hang \\hoffset \\hyphenation \\ialign \\if \\ifcase \
    \\ifcat \\ifdim \\ifeof \\iffalse \\ifhbox \\ifhmode \\ifinner",

    "newline-before : \\ifmmode \\ifnum \\ifodd \\iftrue \\ifundefined \
    \\ifvbox \\ifvmode \\ifvoid \\ifx \\immediate \\input \\leqalignno \
    \\listing \\loop \\magnifiction \\mark \\mathchardef \\message \
    \\narrower \\newbox \\newcount \\newdimen \\newfam \\newhelp",


    /* plain TeX */
    "newline-before : \\newif \\newinsert \\newmuskip \\newlanguage \
    \\newread \\newskip \\newtoks \\newwrite \\noalign \\openin \
    \\openout \\read \\repeat \\show \\showbox \\showboxbreadth \
    \\showdepth \\showhyphens \\showlists \\showthe \\skipdef \
    \\special",


    "newline-before : \\tabalign \\textindent \\toksdef \\tracingall \
    \\tracingcommands \\tracinglostchars \\tracingmacros \
    \\tracingonline \\tracingoutput \\tracingpages \\tracingparagraphs \
    \\tracingrestores \\tracingstats \\vadjust \\valign \\voffset \
    \\vskip \\write \\centerline \\leftline \\line \\parindent \
    \\rightline \\xdef",


    /* plain TeX */
    "standalone : \\allowbreak \\annotations \\batchmode \\bigbreak \
    \\bigskip \\body \\break \\bye \\centering \\dosupereject \\eject \
    \\endletter \\errorstopmode \\filbreak \\frenchspacing \\goodbreak \
    \\indent \\leavevmode \\makelabel \\medbreak \\medskip \\noindent \
    \\nonfrenchspacing \\nonstopmode \\par \\raggedbottom \
    \\raggedcenter \\raggedright \\removelastskip \\scrollmode \
    \\smallbreak \\smallskip \\ttraggedright \\vfil \\vfilneg \\vfill",


    /* TeXinfo 3.2 (NB: this language changes frequently) */
    "chapter-line : @chapter @section @subsection @subsubsection \
    @unnumbered @unnumberedsec @unnumberedsubsec @unnumberedsubsubsec",

    "cite : @cite",

    "command-argument : @code @emph @kbd @key @ref @samp @TeX @var @w \
    @xref",

    "command-line : @author @center @chapheading @clear @deffnx \
    @definfoenclose @defvrx @error @evenfooting @evenheading \
    @everyfooting @everyheading @exdent @footnotestyle @heading \
    @headings @include @majorheading @need @node @oddfooting \
    @oddheading @paragraphindent @set @setchapternewpage @settitle @sp \
    @subheading @subsubheading @subtitle @syncodeindex @synindex \
    @title @top",

    /* TeXinfo indexing commands take args to end of line */
    "command-line : @auindex @cindex @defcodeindex @defindex @findex \
    @kindex @pindex @printindex @shorttitlepage @tindex @vindex",

    "comment : @c @comment",

    /* every other TeXinfo command should be listed here, so that */
    /* out_control_sequence() can check for unrecognized ones that */
    /* may require special handling. */
    "default : @ETC @b @bullet @copyright @dfn @dmn @dots @enddots \
    @equiv @expansion @file @headword @i @math @minus @occur @par \
    @point @print @pxref @r @refill @regularbooksize @result @sc \
    @strong @t @thischapter @thischaptername @thisfile @thispage \
    @thistitle @titlefont @today @value",

    "environment : @cartouche @display @end @flushleft @flushright \
    @format @group @ifinfo @iftex @quotation @smallexample @smalllisp \
    @titlepage",

    "footnote : @footnote",

    "list : @enumerate @ftable @itemize @table @vtable",

    "list-item : @item @itemx",

    "newline-after : @*",

    "newline-before : @appendix @appendixsec @appendixsection \
    @appendixsubsec @appendixsubsubsec @asis @inforef @setfilename \
    @vskip",

    "standalone : @afourpaper @bye @contents @cropmarks @finalout \
    @lowersections @noindent @page @raisesections @shortcontents \
    @smallbook @summarycontents",

    "verbatim : @defcv @deffn @defivar @defmac @defmethod @defop \
    @defopt @defspec @deftp @deftypefn @deftypefun @deftypefunx \
    @deftypevar @deftypevr @defun @defunx @defvar @defvr @example \
    @ignore @lisp @menu @tex",

    /* data terminator */
    (const char*)NULL
};

%}

	/* These settings for lex are about double the required,  */
	/* during initial development.  They can be reduced when  */
	/* the code is stable.  */

%a	 6800
%e	20000
%k	 3000
%n	 2200
%o	 6000
%p	30000

					/* old: ARG	"{"[^{}\n]*"}" */

A		[^{}\n]
					/* not-a-brace-or-newline */

B		"{"{A}*"}"
					/* 1-level braced argument */

C		"{"({A}*{B}*)*"}"
					/* 1,2-level braced argument */

ARG		"{"({A}*{B}*{C}*)*"}"
					/* 1,2,3-level braced argument */

ENVNAME		[A-Za-z][A-Za-z]+[*]?

FLAGNAME	[A-Za-z][A-Za-z0-9]+

LITARG		["][^"]*["]

MACRO		"\\"{ENVNAME}

OPTARG		("["[^\[\]]*"]")?

PLIST		[(][^()]*[)]

PUNCTUATION	[.,:;!?]

W		[ \t]*

%%

"\\$"				|
"\\%"	 			{ out_yytext(); }


"\\\\"[*]?{OPTARG}		{ newline_after(); /* LaTeX */ }


"\\verb*"			{ verb(); /* LaTeX */ }


"\\begin"{W}"{verbatim}"	|
"\\begin"{W}"{verbatim*}"	|
"\\end"{W}"{verbatim}"		|
"\\end"{W}"{verbatim*}"		{ verbatim(); /* LaTeX */ }

"\\begin"{W}"{lstlisting}"	|
"\\end"{W}"{lstlisting}"	{ verbatim(); /* LaTeX */ }

"\\begin"{W}"{tabular}"{W}{ARG} |
"\\begin"{W}"{tabular*}"{W}{ARG} { begin_tabular(); /* LaTeX */ }


"\\end"{W}"{tabular}"		|
"\\end"{W}"{tabular*}"		{ end_tabular(); /* LaTeX */ }


"\\begin"{W}"{displaymath}"	|
"\\["				{ begin_display_math(); /* LaTeX */ }


"\\end"{W}"{displaymath}"	|
"\\]"{PUNCTUATION}?		{ end_display_math(); /* LaTeX */ }


"\\begin"{W}"{math}"		|
"\\("				{ begin_math(); /* LaTeX */ }


"\\end"{W}"{math}"		|
"\\)"{PUNCTUATION}?		{ end_math(); /* LaTeX */ }


"\\begin"{W}"{description}"	|
"\\begin"{W}"{enumerate}"	|
"\\begin"{W}"{itemize}"		|
"\\begin"{W}"{list}"{W}{ARG}{W}{ARG}	|
"\\begin"{W}"{thebibliography}"{W}{ARG}	|
"\\begin"{W}"{trivlist}"	{ begin_list(); /* LaTeX */ }


"\\end"{W}"{description}"	|
"\\end"{W}"{enumerate}"		|
"\\end"{W}"{itemize}"		|
"\\end"{W}"{list}"		|
"\\end"{W}"{thebibliography}"	|
"\\end"{W}"{trivlist}"		{ end_list(); /* LaTeX */ }


"\\bibitem"{W}{OPTARG}{W}{ARG}	|
"\\item"{W}{OPTARG}		{ list_item(); /* LaTeX */ }


"\\begin"{W}"{picture}"{W}{PLIST}({PLIST})?	|
"\\begin"{W}"{"{ENVNAME}"}"{W}{OPTARG}		|
"\\begin"{W}"{"{ENVNAME}"}"{W}{ARG}		|
"\\end"{W}"{"{ENVNAME}"}"	{ environment(); /* LaTeX */ }


"\\bibliography"{W}{ARG}	|
"\\bibliographystyle"{W}{ARG}	{ out_token_on_own_line(yytext); /* LaTeX, LaMSTeX */ }


"\\definecolor"{W}{ARG}		|
"\\enlargethispage"[*]?{W}{ARG}	|
"\\linebreak"{W}{OPTARG}	|
"\\nolinebreak"{W}{OPTARG}	|
"\\nopagebreak"{W}{OPTARG}	|
"\\nopagenumbers"{W}{OPTARG}	|
"\\pagebreak"{W}{OPTARG}	|
"\\pagecolor"{W}{ARG}		|
"\\pagenumbering"{W}{ARG}	|
"\\pagestyle"{W}{ARG}		|
"\\suppressfloats"{W}{OPTARG}	|
"\\thispagestyle"{W}{ARG}	{ out_token_on_own_line(yytext); /* LaTeX */ }


"\\displaybreak"{W}{OPTARG}	{ out_token_on_own_line(yytext); /* AmSLaTeX */ }


"\\let"{W}{MACRO}		{ newline_before(); /* plain TeX */ }


"\\obeylines"			|
"\\obeyspaces"			{ obey(); /* plain TeX */ }


"\\obeywhitespace"		{ obey(); /* ETeX */ }


"\\widestnumber"{W}{MACRO}	{ newline_before(); /* AmSTeX */ }


"\\ChangeBuffer"{W}{ARG}	|
"\\MultLineGap"{W}{ARG}		|
"\\ResetBuffer"{W}{ARG}		|
"\\captionwidth"{W}{ARG}	|
"\\hcorrection"{W}{ARG}		|
"\\multlinegap"{W}{ARG}		|
"\\operatorname"{W}{ARG}	|
"\\pageheight"{W}{ARG}		|
"\\pagewidth"{W}{ARG}		|
"\\spreadlines"{W}{ARG}		|
"\\spreadmatrixlines"{W}{ARG}	|
"\\vcorrection"{W}{ARG}		{ out_token_on_own_line(yytext); /* AmSTeX */ }


"\\item"{W}{LITARG}		|
"\\itemitem"{W}{ARG}		{ list_item(); /* AmSTeX */ }


"\\Offset"{W}{MACRO}		|
"\\Reset"{W}{MACRO}		|
"\\newHL"{W}{MACRO}		|
"\\newfontstyle"{W}{MACRO}	|
"\\newhl"{W}{MACRO}		|
"\\newnumstyle"{W}{MACRO}	|
"\\newpost"{W}{MACRO}		|
"\\newpre"{W}{MACRO}		|
"\\newstyle"{W}{MACRO}		|
"\\newword"{W}{MACRO}		{ newline_before(); /* LAmSTeX */ }


"@*"				{	/* TeXinfo */
				    if (in_texinfo)
					newline_after();
				    else
					out_yytext();
				}


@[@{}]				{ out_yytext(); /* TeXinfo */ }


[$]{PUNCTUATION}?		{ math(); /* any TeX */ }


[$][$]{PUNCTUATION}?		{ display_math(); /* any TeX */ }


[.:!?][ \t]+			{ /* preserve double space at end of sentence */
				    out_char((int)yytext[0]);
				    out_blank();
				    out_blank();
				}


[ \t]+				{ blank(); }


^[ \t]*[%].*			|
[%].*				{ out_comment(); }


"\\\r\n"			|
"\\\n"				{ line_number++; out_verbatim_string("\\ %\n"); }


@itemize[ \t]+.*$		|
@[fv]?table[ \t]+@[A-Za-z]+	{
				    if (in_texinfo)
				    {
					normalize_space(yytext);
					begin_list();
				    }
				    else
					out_yytext();
				}



@ifset[ \t]+{FLAGNAME}		|
@ifclear[ \t]+{FLAGNAME}	{
				    if (in_texinfo)
					begin_environment();
				    else
					out_yytext();
				}


@end[ \t]+[A-Za-z]+		{
				    if (in_texinfo)
				    {
					normalize_space(yytext);
					end_environment();
				    }
				    else
					out_yytext();
				}


@[A-Za-z]+			{
				    if (in_texinfo)
					out_control_sequence();
				    else
					out_yytext();
				}


[\\][@A-Za-z]+			{ out_control_sequence(); }


"\\{"				|
"\\}"				|
"\\&"				|
[\\][^@A-Za-z\\\r\n]		|
[\\][@A-Za-z]+[0-9]*[@A-Za-z]+	|
[A-Za-z0-9]+			|
[0-9]+[.][0-9]+			|
[0-9]+[.]			|
[.][0-9]+			|
[^ \t\f\r\n\013@\\_^$&\[\]{}()%]+	{ out_yytext(); }


[\f\r\t\013]			{ out_blank(); }	/* other whitespace */


\n				{ line_number++; line_end(); }


"&"				{ ampersand(); }


"{"				{ begin_brace(); }


"}"				{ end_brace(); }


"["				{ begin_bracket(); }


"]"				{ end_bracket(); }


"("				{ begin_parenthesis(); }


")"				{ end_parenthesis(); }


.				{ out_char((int)yytext[0]); }


%%

#if HAVE_STRNICMP
#else
/* Put this one first, so that we do not have to supply a (possibly-conflicting)
 prototype */
#define strnicmp Strnicmp /* make name private to avoid library conflicts */
int
#if STDC
strnicmp(
const char	*s1,
const char	*s2,
size_t		n
)
#else /* NOT STDC */
strnicmp(s1,s2,n)
const char	*s1;
const char	*s2;
size_t		n;
#endif /* STDC */
{
    int	   c1;
    int	   c2;
    int	   result;

    for (; (n > 0) && *s1 && *s2; ++s1, ++s2, --n)
    {
	c1 = 0xff & (islower(*s1) ? (int)*s1 : tolower(*s1));
	c2 = 0xff & (islower(*s2) ? (int)*s2 : tolower(*s2));
	if (c1 < c2)
	    return (-1);
	else if (c1 > c2)
	    return (1);
    }
    if (n <= 0)		   /* first n characters match */
	result = 0;
    else if (*s1 == '\0')
	result = ((*s2 == '\0') ? 0 : -1);
    else /* (*s2 == '\0') */
	result = 1;

    return (result);
}
#endif /* HAVE_STRNICMP */


#if STDC
static void
add_style(const char *style, const char *name)
#else
static void
add_style(style,name)
const char *style;
const char *name;
#endif
{
    Style *s;
    Style temp;

#if !defined(STYLECHUNK)
#define STYLECHUNK		900	/* big enough for builtin table */
#endif

    if (style_table.size >= style_table.max_size)
    {					/* create or enlarge table */
	style_table.max_size += STYLECHUNK;
	style_table.table =
	    (Style*)get_memory(style_table.table,
			       sizeof(Style) * style_table.max_size);
    }
    if (strcmp("default",style) == 0)
    {
	s = &temp;
	s->style = "default";
	s->action = Is_Control_Sequence(name) ? out_yytext : environment;
    }
    else
	s = get_style_by_style(style);

    if (s == (Style*)NULL)		/* should never happen */
    {

#undef MSG_FMT
#define MSG_FMT "unrecognized style %.39s command %.39s in style file: ignored"

	char tempmsg[sizeof(MSG_FMT) + 39 + 39 + 1];

	SPRINTF(tempmsg, MSG_FMT, style, name);
	warning(tempmsg);
    }
    else
    {
	style_table.table[style_table.size].name = (const char*)dupstr(name);
	style_table.table[style_table.size].style = s->style;
	style_table.table[style_table.size].action = s->action;
	style_table.size++;
    }
}


#if STDC
static void
adjust_brace_level(int n)
#else
static void
adjust_brace_level(n)
int n;
#endif
{
    if ((n < 0) && (brace_level < MAXDELIMITERLEVEL))
	brace_line[brace_level] = 0;
    brace_level += n;
    if (brace_level < 0)
	check_brace_level_is_zero("this line",1);
    if ((n > 0) && (brace_level < MAXDELIMITERLEVEL))
	brace_line[brace_level] = line_number;
}


#if STDC
static void
adjust_bracket_level(int n)
#else
static void
adjust_bracket_level(n)
int n;
#endif
{
    if ((n < 0) && (bracket_level < MAXDELIMITERLEVEL))
	bracket_line[bracket_level] = 0;
    bracket_level += n;
    if (bracket_level < 0)
	check_bracket_level_is_zero("this line",1);
    if ((n > 0) && (bracket_level < MAXDELIMITERLEVEL))
	bracket_line[bracket_level] = line_number;
}


#if STDC
void
adjust_environment_level(int n, const char *command)
#else
void
adjust_environment_level(n,command)
int n;
const char *command;
#endif
{
    if ((n < 0) && (environment_level < MAXENVIRONMENTLEVEL))
    {
	check_environment_name_match(environment_name[environment_level],
					  get_name(command));
	if (environment_name[environment_level] != (char*)NULL)
	{
	    free((VOIDP)environment_name[environment_level]);
	    environment_name[environment_level] = (char*)NULL;
	}
	environment_line[environment_level] = 0;
    }
    environment_level += n;
    if (environment_level < 0)
	check_environment_level_is_zero("this line",1);
    if ((n > 0) && (environment_level < MAXENVIRONMENTLEVEL))
    {
	environment_line[environment_level] = line_number;
	if (environment_name[environment_level] != (char*)NULL)
	    free((VOIDP)environment_name[environment_level]);
	environment_name[environment_level] = dupstr(get_name(command));
    }
}


#if STDC
static void
adjust_list_level(int n, const char *command)
#else
static void
adjust_list_level(n,command)
int n;
const char *command;
#endif
{
    command = (const char*)NULL;	/* to suppress compiler warnings */
					/* about unused argument */
    environment_level += n;
}


#if STDC
static void
adjust_math_level(int n)
#else
static void
adjust_math_level(n)
int n;
#endif
{
    if ((n < 0) && (math_level < MAXENVIRONMENTLEVEL))
	math_line[math_level] = 0;
    math_level += n;
    if (math_level < 0)
	check_math_level_is_zero("this line",1);
    if ((n > 0) && (math_level < MAXENVIRONMENTLEVEL))
	math_line[math_level] = line_number;
    if (math_level > 1)
    {
	warning("illegal nested math environments");
	FPRINTF(stderr, "\tUnclosed math environments at lines");
	out_linenumbers(math_line,(const char**)NULL,
			MIN(math_level,MAXENVIRONMENTLEVEL-1));
    }
}


#if STDC
static void
adjust_parenthesis_level(int n)
#else
static void
adjust_parenthesis_level(n)
int n;
#endif
{
    if ((n < 0) && (parenthesis_level < MAXDELIMITERLEVEL))
	parenthesis_line[parenthesis_level] = 0;
    parenthesis_level += n;
    if (parenthesis_level < 0)
	check_parenthesis_level_is_zero("this line",1);
    if ((n > 0) && (parenthesis_level < MAXDELIMITERLEVEL))
	parenthesis_line[parenthesis_level] = line_number;
}


static void
ampersand(VOID_ARG)
{
    if (in_tabular)		/* line up tabular ampersands on tab columns */
	out_leading_blanks(((line_length() + 7) & ~7) - line_length());
    out_yytext();
}


static void
author(VOID_ARG)
{
    PUTSTDERR("Author:\n");
    PUTSTDERR("\tNelson H. F. Beebe\n");
    PUTSTDERR("\tCenter for Scientific Computing\n");
    PUTSTDERR("\tUniversity of Utah\n");
    PUTSTDERR("\tDepartment of Mathematics, 322 INSCC\n");
    PUTSTDERR("\t155 S 1400 E RM 233\n");
    PUTSTDERR("\tSalt Lake City, UT 84112-0090\n");
    PUTSTDERR("\tUSA\n");
    PUTSTDERR("\tEmail: beebe@math.utah.edu, beebe@acm.org, beebe@ieee.org (Internet)\n");
    PUTSTDERR("\tWWW URL: http://www.math.utah.edu/~beebe\n");
    PUTSTDERR("\tTel: +1 801 581 5254\n");
    PUTSTDERR("\tFAX: +1 801 585 1640, +1 801 581 4148\n");
}


#if STDC
static void
bad_option(const char *option)
#else
static void
bad_option(option)
const char *option;
#endif
{

#undef MSG_FMT
#define MSG_FMT "unrecognized option [%.40s] ignored"

    char tempmsg[sizeof(MSG_FMT) + 40 + 1];

    SPRINTF(tempmsg, MSG_FMT, option);
    warning(tempmsg);
    usage();
    exit(EXIT_FAILURE);
}


static void
begin_brace(VOID_ARG)
{
    adjust_brace_level(1);
    out_yytext();
}


static void
begin_bracket(VOID_ARG)
{
    adjust_bracket_level(1);
    out_yytext();
}


static void
begin_display_math(VOID_ARG)
{
    remove_separating_blanks();
    adjust_math_level(1);
    out_token_on_own_line((default_begin_display_math == (const char*)NULL) ?
			  yytext : default_begin_display_math);
    adjust_environment_level(1,yytext);
}


static void
begin_environment(VOID_ARG)
{
    if (!Is_TeXinfo(yytext))
	remove_separating_blanks();

    out_token_on_own_line(yytext);

    /* Special cases: environment level is not altered by
       AmSTeX: \topmatter \endtopmatter and \document \enddocument
       LaTeX: \begin{document} \end{document}
    */

    if ((strcmp("\\begin{document}", yytext) == 0) ||
	(strcmp("\\document", yytext) == 0) ||
	(strcmp("\\topmatter", yytext) == 0))
	NOOP;
    else
	adjust_environment_level(1,yytext);
}


static void
begin_list(VOID_ARG)
{
    if (!Is_TeXinfo(yytext))		/* protect "@table @kbd" et al */
	remove_separating_blanks();
    out_token_on_own_line(yytext);
    adjust_environment_level(2,yytext);
}


static void
begin_math(VOID_ARG)
{
    remove_separating_blanks();
    adjust_math_level(1);
    if (display_math_mode)
    {
	out_token_on_own_line((default_begin_math == (const char*)NULL) ?
			      yytext : default_begin_math);
	adjust_environment_level(1,yytext);
    }
    else
    {
	out_string(default_begin_math == (const char*)NULL ?
		   yytext : default_begin_math);
	out_blank();
    }
}


static void
begin_parenthesis(VOID_ARG)
{
    adjust_parenthesis_level(1);
    out_yytext();
}


static void
begin_tabular(VOID_ARG)
{
    remove_separating_blanks();

    if (tabular_like_verbatim)
    {
	out_linebreak();
	trim_line(0);	/* want \begin{tabular} and \end{tabular} on new line */
	out_verbatim_string((const char*)&yytext[0]);
	copy_verbatim();  /* handles \end{tabular} and \end{tabular*} */
    }
    else
    {			/* keep \begin{tabular}{longarg} on one line */
	out_verbatim_string((const char*)&yytext[0]);
	adjust_environment_level(1,yytext);
	in_tabular = 1;
	copy_tabular();  /* handles \end{tabular} and \end{tabular*} */
	check_brace_level_is_zero("end of tabular",0);
	check_bracket_level_is_zero("end of tabular",0);
	check_parenthesis_level_is_zero("end of tabular",0);
    }
}


static void
begin_verbatim(VOID_ARG)
{
    adjust_environment_level(1,yytext);
    out_linebreak();
    trim_line(0);	/* want \begin{verbatim} and \end{verbatim} on new line */
    out_verbatim_string((const char*)&yytext[0]);
}


static void
begin_verbatim_output(VOID_ARG)
{
    in_verbatim = 1;
}


static void
blank(VOID_ARG)
{
    int c_last;
    int c_lastm1;
    int c_lastm2;

    delete_spaces();

    c_last = last_char(0);
    c_lastm1 = last_char(-1);
    c_lastm2 = last_char(-2);

    out_blank();		/* collapse multiple spaces into single one */

    switch (c_last)
    {
    case '.':
	if (isupper(c_lastm1))
	{
	    if (isspace(c_lastm2) || (c_lastm2 == '~'))
		break;			/* looks like [ ~][A-Z][.] */
	    else
		out_blank();		/* assume sentence end */
	}
	else if ((c_lastm2 == '.') && isalpha(c_lastm1))
	    break;			/* looks like "e.g." or "i.e." */
	else
	    out_blank();		/* assume sentence end */
	break;

    case ':':
    case '!':
    case '?':
	out_blank();		/* two spaces after these punctuators */
	break;

    default:				/* no additional space */
	break;
    }
}


#if STDC
static void
check_brace_level_is_zero(const char *msg, int reset)
#else
static void
check_brace_level_is_zero(msg, reset)
const char *msg;
int reset;
#endif
{
    if (brace_level != 0)
    {

#undef MSG_FMT
#define MSG_FMT "%.8s brace level at %.80s%s"
#undef MSG_RESET
#define MSG_RESET ": reset to zero"

	char tempmsg[sizeof(MSG_FMT) + 8 + 80 + sizeof(MSG_RESET) + 1];

	SPRINTF(tempmsg, MSG_FMT, (brace_level < 0) ? "negative" : "positive",
		msg, reset ? MSG_RESET : "");
	warning(tempmsg);
	if (brace_level > 0)
	{
	    FPRINTF(stderr, "\tUnclosed open brace%s at line%s",
		    (brace_level > 1) ? "s" : "",
		    (brace_level > 1) ? "s" : "");
	    out_linenumbers(brace_line,(const char**)NULL,
			    MIN(brace_level,MAXDELIMITERLEVEL-1));
	}
	if (reset)
	{
	    brace_level = 0;
	    brace_line[brace_level] = 0;
	}
    }
}


#if STDC
static void
check_bracket_level_is_zero(const char *msg, int reset)
#else
static void
check_bracket_level_is_zero(msg, reset)
const char *msg;
int reset;
#endif
{
    if (bracket_level != 0)
    {

#undef MSG_FMT
#define MSG_FMT "%.8s bracket level at %.80s%s"
#undef MSG_RESET
#define MSG_RESET ": reset to zero"

	char tempmsg[sizeof(MSG_FMT) + 8 + 80 + 1];

	SPRINTF(tempmsg, MSG_FMT,
		(bracket_level < 0) ? "negative" : "positive", msg,
		reset ? MSG_RESET : "");
	warning(tempmsg);
	if (bracket_level > 0)
	{
	    FPRINTF(stderr, "\tUnclosed open bracket%s at line%s",
		    (bracket_level > 1) ? "s" : "",
		    (bracket_level > 1) ? "s" : "");
	    out_linenumbers(bracket_line,(const char**)NULL,
			    MIN(bracket_level,MAXDELIMITERLEVEL-1));
	}
	if (reset)
	{
	    bracket_level = 0;
	    bracket_line[bracket_level] = 0;
	}
    }
}


#if STDC
static void
check_delimiter_level(int c)
#else
static void
check_delimiter_level(c)
int c;
#endif
{
    switch (c)
    {
    case '{':	adjust_brace_level(1); break;
    case '}':	adjust_brace_level(-1); break;
    case '[':	adjust_bracket_level(1); break;
    case ']':	adjust_bracket_level(-1); break;
    case '(':	adjust_parenthesis_level(1); break;
    case ')':  	adjust_parenthesis_level(-1); break;
    case '$':   if (!in_texinfo)
		    adjust_math_level((math_level > 0) ? -1 : 1);
		break;
    default:	NOOP; break;
    }
}


#if STDC
static void
check_environment_level_is_zero(const char *msg, int reset)
#else
static void
check_environment_level_is_zero(msg, reset)
const char *msg;
int reset;
#endif
{
    if (environment_level != 0)
    {

#undef MSG_FMT
#define MSG_FMT "%.8s environment level at %.80s%s"
#undef MSG_RESET
#define MSG_RESET ": reset to zero"

	char tempmsg[sizeof(MSG_FMT) + 8 + 80 + 1];

	SPRINTF(tempmsg, MSG_FMT,
		(environment_level < 0) ? "negative" : "positive", msg,
		reset ? MSG_RESET : "");
	warning(tempmsg);
	if (environment_level > 0)
	{
	    FPRINTF(stderr, "\tUnclosed open environment%s at line%s",
		    (environment_level > 1) ? "s" : "",
		    (environment_level > 1) ? "s" : "");
	    out_linenumbers(environment_line,(const char**)&environment_name[0],
			    MIN(environment_level,MAXENVIRONMENTLEVEL-1));
	}
	if (reset)
	{
	    environment_level = 0;
	    environment_line[environment_level] = 0;
	    if (environment_name[environment_level] != (char*)NULL)
		free((VOIDP)environment_name[environment_level]);
	    environment_name[environment_level] = (char*)NULL;
	}
    }
}


#if STDC
static void
check_environment_name_match(const char *begin, const char *end)
#else
static void
check_environment_name_match(begin,end)
const char *begin;
const char *end;
#endif
{
    int match;

#define MATCH(a,b)	(strcmp((a),(b)) == 0)

    if (begin == (const char*)NULL)
	begin = "";
    if (end == (const char*)NULL)
	end = "";

    if (Is_Control_Sequence_Character(begin[0]))
	match = MATCH(begin,end);
    else if (MATCH(begin,"\\["))
	match = MATCH(end,"\\]");
    else if (MATCH(begin,"\\("))
	match = MATCH(end,"\\)");
    else if (MATCH(begin,"$$"))
	match = MATCH(end,"$$");
    else if (MATCH(begin,"$"))
	match = MATCH(end,"$");
    else if (MATCH(begin,""))		/* if called from out_string() */
	match = MATCH(end,"");
    else
	match = 0;

    if (!match)
    {
	warning("begin/end environment name mismatch");
	FPRINTF(stderr, "\tbegin {%s} at line %d: end {%s} at line %d\n",
		begin, environment_line[environment_level],
		end, line_number);
    }
}


#if STDC
static void
check_math_level_is_zero(const char *msg, int reset)
#else
static void
check_math_level_is_zero(msg, reset)
const char *msg;
int reset;
#endif
{
    if (math_level != 0)
    {

#undef MSG_FMT
#define MSG_FMT "positive math mode level at %.80s%s"
#undef MSG_RESET
#define MSG_RESET ": reset to zero"

	char tempmsg[sizeof(MSG_FMT) + 80 + 1];

	SPRINTF(tempmsg, MSG_FMT, msg, reset ? MSG_RESET : "");
	warning(tempmsg);
	if (math_level > 0)
	{
	    FPRINTF(stderr, "\tUnclosed math environment%s at line%s",
		    (math_level > 1) ? "s" : "",
		    (math_level > 1) ? "s" : "");
	    out_linenumbers(math_line,(const char**)NULL,
			    MIN(math_level,MAXENVIRONMENTLEVEL-1));
	}
	if (reset)
	{
	    math_level = 0;
	    math_line[math_level] = 0;
	}
    }
}


#if STDC
static void
check_parenthesis_level_is_zero(const char *msg, int reset)
#else
static void
check_parenthesis_level_is_zero(msg, reset)
const char *msg;
int reset;
#endif
{
    if (parenthesis_level != 0)
    {

#undef MSG_FMT
#define MSG_FMT "%.8s parenthesis level at %.80s%s"
#undef MSG_RESET
#define MSG_RESET ": reset to zero"

	char tempmsg[sizeof(MSG_FMT) + 8 + 80 + 1];

	SPRINTF(tempmsg, MSG_FMT,
		(parenthesis_level < 0) ? "negative" : "positive",
		msg, reset ? MSG_RESET : "");
	warning(tempmsg);
	if (parenthesis_level > 0)
	{
	    FPRINTF(stderr, "\tUnclosed open parenthes%s at line%s",
		    (parenthesis_level > 1) ? "es" : "is",
		    (parenthesis_level > 1) ? "s" : "");
	    out_linenumbers(parenthesis_line, (const char**)NULL,
			    MIN(parenthesis_level,MAXDELIMITERLEVEL-1));
	}
	if (reset)
	{
	    parenthesis_level = 0;
	    parenthesis_line[parenthesis_level] = 0;
	}
    }
}


static void
cite(VOID_ARG)
{
    int c_last;

    trim_line(0);
    if (line_length() > 0)
    {
	c_last = last_char(0);
	if (c_last == '~') /* delete undesirable ties before \cite and \nocite */
	    next_position--;
	out_newline();
    }
    out_yytext();
}


static void
command_argument(VOID_ARG)
{
    out_yytext();
    copy_argument();
}


static void
command_line(VOID_ARG)
{
    newline_before();
    copy_to_end_of_line();
}


static void
comment_and_newline_before(VOID_ARG)
{
    int c_last;

    trim_line(0);
    if (line_length() > 0)
    {
	c_last = last_char(0);
	if (c_last != '%')
	    out_char((int)'%');
	out_newline();
    }
    out_yytext();
}


static void
copy_argument(VOID_ARG)
{
    register int c;
    register int level;

    for (c = input_char(); isspace(c) && !LEXEOF(c); c = input_char())
	NOOP;		/* discard space between command and argument */

    if (c != '{')			/* '}' for balance */
    {

#undef MSG_FMT
#define MSG_FMT "expected braced argument after `%.39s'"

	char tempmsg[sizeof(MSG_FMT) + 39 + 1];

	unput_char(c);
	SPRINTF(tempmsg, MSG_FMT, yytext);
	warning(tempmsg);
	return;
    }

    unput_char(c);
    level = 0;
    do
    {
	c = input_char();
	switch (c)
	{
	case '\n':
	    if (Is_TeXinfo(yytext))
		c = ' ';
	    else
	    {

#undef MSG_FMT
#define MSG_FMT "unexpected newline inside braced argument of `%.39s'"

		char tempmsg[sizeof(MSG_FMT) + 39 + 1];

		SPRINTF(tempmsg, MSG_FMT, yytext);
		warning(tempmsg);
		level = 0;		/* force early bailout to prevent */
					/* rest of file being collected as */
					/* an argument */
	    }
	    break;

	case '@':
	    out_char(c);		/* TeXinfo escape sequence: @@ et al */
	    c = input_char();
	    break;

	case '{':			/* '}' for balance */
	    level++;
	    break;
					/* '{' for balance */
	case '}':
	    level--;
	    break;

	default:
	    break;
	}
	out_char(c);
    } while ((level > 0) && !LEXEOF(c));
}


static void
copy_to_end_of_line(VOID_ARG)
{
    register int c;

    begin_verbatim_output();
    for (c = input_char(); c != '\n' && !LEXEOF(c); c = input_char())
	out_char(c);
    out_char('\n');
    end_verbatim_output();
}


static void
copy_verbatim(VOID_ARG)
{
    /* Copy up to matching \end{tabular}, \end{tabular*}, \end{verbatim},
       or \end{verbatim*} */

    register int c;
    static char *endcommand = (char*)NULL;
    static size_t len_endcommand = 0;
    size_t len;
    register int matches;
    char *p;

    if (Is_TeXinfo(yytext))		/* expect @end name */
	len = strlen(yytext) + sizeof("end ") + 1;
    else				/* expect \endxxx */
	len = strlen(yytext) - sizeof("begin") + sizeof("end") + 1;

    if (len > len_endcommand)
    {					/* create or enlarge buffer */
	endcommand = (char*)get_memory(endcommand, len);
	len_endcommand = len;
    }
    if (Is_TeXinfo(yytext))		/* expect @end name */
    {
	(void)strcpy(endcommand,"@end ");
	(void)strcpy(&endcommand[5], (const char*)&yytext[1]);
    }
    else				/* expect \end{verbatim} or \endxxx */
    {
	(void)strcpy(endcommand,"\\end");
	(void)strcpy(&endcommand[4], (const char*)&yytext[6]);
    }

    p = strchr(endcommand,'}');		/* remove any second argument */
    if (p != (char*)NULL)
	p[1] = '\0';

    begin_verbatim_output();

    matches = 0;
    while ((c = input_char(), !LEXEOF(c)))
    {
	if (endcommand[matches] == '\0')
	{
	    unput_char(c);
	    delete_indentation();
	    adjust_environment_level(-1,endcommand);
	    out_verbatim_char('\n');
	    end_verbatim_output();
	    return;			/* here's the loop exit */
	}
	out_verbatim_char(c);
	if (endcommand[matches] == c)
	    matches++;
	else
	    matches = 0;
    }

    end_verbatim_output();
}


static void
copyright(VOID_ARG)
{
    FPRINTF(stderr,"########################################################################\n");
    FPRINTF(stderr,"########################################################################\n");
    FPRINTF(stderr,"########################################################################\n");
    FPRINTF(stderr,"###                                                                  ###\n");
    FPRINTF(stderr,"### texpretty: prettyprint AmSTeX, AmsLaTeX, ETeX, LAmSTeX, LaTeX,   ###\n");
    FPRINTF(stderr,"###             LaTeXinfo, plain TeX, SliTeX, and TeXinfo files      ###\n");
    FPRINTF(stderr,"###                                                                  ###\n");
    FPRINTF(stderr,"###              Copyright (C) 1999 Nelson H. F. Beebe               ###\n");
    FPRINTF(stderr,"###                                                                  ###\n");
    FPRINTF(stderr,"### This program is covered by the GNU General Public License (GPL), ###\n");
    FPRINTF(stderr,"### version 2 or later, available as the file COPYING in the program ###\n");
    FPRINTF(stderr,"### source distribution, and on the Internet at                      ###\n");
    FPRINTF(stderr,"###                                                                  ###\n");
    FPRINTF(stderr,"###               ftp://ftp.gnu.org/gnu/GPL                          ###\n");
    FPRINTF(stderr,"###                                                                  ###\n");
    FPRINTF(stderr,"###               http://www.gnu.org/copyleft/gpl.html               ###\n");
    FPRINTF(stderr,"###                                                                  ###\n");
    FPRINTF(stderr,"### This program is free software; you can redistribute it and/or    ###\n");
    FPRINTF(stderr,"### modify it under the terms of the GNU General Public License as   ###\n");
    FPRINTF(stderr,"### published by the Free Software Foundation; either version 2 of   ###\n");
    FPRINTF(stderr,"### the License, or (at your option) any later version.              ###\n");
    FPRINTF(stderr,"###                                                                  ###\n");
    FPRINTF(stderr,"### This program is distributed in the hope that it will be useful,  ###\n");
    FPRINTF(stderr,"### but WITHOUT ANY WARRANTY; without even the implied warranty of   ###\n");
    FPRINTF(stderr,"### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    ###\n");
    FPRINTF(stderr,"### GNU General Public License for more details.                     ###\n");
    FPRINTF(stderr,"###                                                                  ###\n");
    FPRINTF(stderr,"### You should have received a copy of the GNU General Public        ###\n");
    FPRINTF(stderr,"### License along with this program; if not, write to the Free       ###\n");
    FPRINTF(stderr,"### Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,   ###\n");
    FPRINTF(stderr,"### MA 02111-1307 USA                                                ###\n");
    FPRINTF(stderr,"###                                                                  ###\n");
    FPRINTF(stderr,"### The World-Wide Web master archive for this program is located at ###\n");
    FPRINTF(stderr,"###                                                                  ###\n");
    FPRINTF(stderr,"###     ftp://ftp.math.utah.edu/pub/misc/texpretty-x.yz.tar.gz       ###\n");
    FPRINTF(stderr,"###     http://www.math.utah.edu/pub/misc/texpretty-x.yz.tar.gz      ###\n");
    FPRINTF(stderr,"###                                                                  ###\n");
    FPRINTF(stderr,"### where x.yz is a version number.                                  ###\n");
    FPRINTF(stderr,"###                                                                  ###\n");
    FPRINTF(stderr,"########################################################################\n");
    FPRINTF(stderr,"########################################################################\n");
    FPRINTF(stderr,"########################################################################\n");
}


static void
delete_indentation(VOID_ARG)
{   /* delete indentation from line_buffer[] and move the remaining text left */
    register int k;

    for (k = 0; (k < next_position) && (line_buffer[k] == ' '); ++k)
	NOOP;
    if (k > 0)
    {
	for (next_position = 0; line_buffer[k]; )
	    line_buffer[next_position++] = line_buffer[k++];
	line_buffer[next_position] = '\0';
    }
}


void
delete_spaces(VOID_ARG)
{
    trim_line(indentation_size());
}


static void
display_math(VOID_ARG)
{
    if (in_texinfo)
	out_yytext();
    else if (math_level > 0)
	end_display_math();
    else
	begin_display_math();
}


#if STDC
static void
do_macro_arguments(const char *pattern)
#else
static void
do_macro_arguments(pattern)
const char *pattern;
#endif
{
    /*******************************************************************
    Copy macro arguments from the input stream to the output stream in
    verbatim mode, discarding interargument spaces.  The argument
    pattern looks something like these:

	\name*
	\name()
	\name[]{}
	\name""
	\name{}
	\name*[]""{}
	\name\

    Here *, [], and "" represent optional arguments, () and {} represent
    required arguments, and a final backslash stands for another
    arbitrary control sequence.  * stands for itself, [] for a
    balanced-bracket sequence, "" for a quoted sequence, {} for a
    balanced-brace sequence, and () for a balanced-parenthesis sequence.
    All other characters, including space, are illegal in patterns.

    The initial \name part of the pattern is simply a placeholder that
    is ignored.

    AmSLaTeX and LaTeX use * after certain control sequences to
    represent variants, and use [] for optional arguments.  AmSTeX and
    LAmSTeX use "" for optional arguments.  All use {} for required
    arguments.
    *******************************************************************/

    const char *original_pattern;

    original_pattern = pattern;		/* in case of warning messages */

    if (Is_Control_Sequence(pattern))
	++pattern;			/* move past escape character */
    while (Is_Control_Sequence_Character(*pattern))
	++pattern;			/* move past control sequence name */

    begin_verbatim_output();
    for (NOOP; *pattern; ++pattern)
    {
	register int c;
	register int level;
	int open_delimiter;
	int close_delimiter;

	while ((c = input_char()) && !LEXEOF(c) && isspace(c))
	    NOOP;			/* ignore inter-arg space */

	if (LEXEOF(c))
	    break;

	switch (*pattern)		/* match against the pattern */
	{
	case '\\':			/* expect control sequence name */
	    if (c == '\\')
	    {
		out_verbatim_char(' ');	/* supply separating space */
		out_verbatim_char(c);
		for (;;)
		{
		    c = input_char();
		    if (LEXEOF(c))
			break;
		    else if (Is_Control_Sequence_Character(c))
			out_verbatim_char(c);
		    else
			break;
		}
	    }
	    unput_char(c);
	    break;

	case '*':
	    if (c == '*')
		out_verbatim_char(c);
	    else
		unput_char(c);
	    break;

	case '(':			/* ')' for balance */
	    open_delimiter = '(';
	    close_delimiter = ')';
	    goto BALANCED_DELIMITER;

	case '[':			/* ']' for balance */
	    open_delimiter = '[';
	    close_delimiter = ']';
	    goto BALANCED_DELIMITER;

	case '{':			/* '}' for balance */
	    open_delimiter = '{';
	    close_delimiter = '}';
	    goto BALANCED_DELIMITER;

	BALANCED_DELIMITER:
	    if (*++pattern != close_delimiter)	/* expect (), [], or {} */
		goto BADPATTERN;
	    if (c == open_delimiter)
	    {
		out_verbatim_char(c);
		for (level = 1; level > 0; NOOP)
		{
		    c = input_char();
		    if (LEXEOF(c))
			break;
		    out_verbatim_char(c);
		    if (c == open_delimiter)
			level++;
		    else if (c == close_delimiter)
			level--;
		}
	    }
	    else
		unput_char(c);
	    break;

	case '"':
	    if (*++pattern != '"')	/* expect "" */
		goto BADPATTERN;
	    if (c == '"')		/* copy quoted string */
	    {
		out_verbatim_char(c);
		do
		{
		    c = input_char();
		    if (LEXEOF(c))
			break;
		    out_verbatim_char(c);
		} while (c != '"');
	    }
	    else
		unput_char(c);
	    break;

	default:
	BADPATTERN:
	    {

#undef MSG_FMT
#define MSG_FMT "unexpected character `%c' in pattern `%.39s'"

		char tempmsg[sizeof(MSG_FMT) + 39 + 1];

		unput_char(c);
		SPRINTF(tempmsg, MSG_FMT, *pattern,original_pattern);
		end_verbatim_output();
		return;
	    }
	}
    }
    end_verbatim_output();
}


#if STDC
static void
do_nothing(const char *pattern)
#else
static void
do_nothing(pattern)
const char *pattern;
#endif
{
			/* post-outout do-nothing action for out_string() */
}


#if STDC
static void
do_style_file(const char *the_filename, int warn_if_no_file)
#else
static void
do_style_file(the_filename, warn_if_no_file)
const char *the_filename;
int warn_if_no_file;
#endif
{
    FILE *fpstyle;
    int style_line_number;

    fpstyle = fopen(the_filename,"r");
    if (fpstyle == (FILE*)NULL)
    {
	if (warn_if_no_file)
	{

#undef MSG_FMT
#define MSG_FMT "unable to open style file %.255s"

	    char tempmsg[sizeof(MSG_FMT) + 255 + 1];

	    SPRINTF(tempmsg, MSG_FMT, the_filename);
	    warning(tempmsg);
	}
    }
    else
    {
	char line[MAXBUF+2];		/* space for two trailing NULs */

	style_line_number = 0;
	while (fgets(line, sizeof(line), fpstyle) != (char*)NULL)
	{
	    style_line_number++;
	    do_style_line(the_filename, style_line_number, line);
	}
    }
}


#if STDC
static void
do_style_line(const char *source, int style_line_number, char *line)
#else
static void
do_style_line(source, style_line_number, line)
const char *source;
int style_line_number;
char *line;
#endif
{
    int n;
    char overwritten_char;
    char *p;
    const char *token[2];

#define STYLECHARS "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-"
#define TOKENCHARS "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ\\@:*()[]{}"

    p = strchr(line,'%');
    if (p == (char*)NULL)
	p = strchr(line,'\0');
    p[0] = '\0';			/* need two consecutive terminators */
    p[1] = '\0';

    for (p = strchr(line,'\0') - 1; (p > line) && isspace(*p); --p)
	*p = '\0';			/* strip trailing space */

    if (p <= line)
	return;				/* ignore empty lines */

    token[1] = "";
    for (n = 0, p = get_token(line,STYLECHARS,&overwritten_char);
	 (p != (char*)NULL) && *p;
	 p = get_token(p + (int)strlen(p) + 1, TOKENCHARS,
			       &overwritten_char))
    {
	/* In the interests of storage economy, we tolerate some */
	/* ugliness here: we want to be able to store the tokens */
	/* in the line[] buffer, and we want normal C NUL string */
	/* terminators.   If there is no space around the colon */
	/* separating the style class from the macro and */
	/* environment names, there is no room for the NULs, */
	/* so we check for this, and store a constant colon */
	/* string instead. */

	if (n == 0)
	{
	    token[n++] = p;
	    if (overwritten_char == ':') /* no space before colon */
		token[n++] = ":";
	}
	else if (n == 1)
	    token[n++] = p;
	else if (strcmp(token[1],":") == 0)
	    add_style(token[0], p);
    }

    if (strcmp(token[1],":") != 0)
    {

#undef MSG_FMT
#define MSG_FMT "expected style : name name ... in %.80s:%d:%.255s"

	char tempmsg[sizeof(MSG_FMT) + 80 + 18 + 255 + 1];

	SPRINTF(tempmsg, MSG_FMT, source, style_line_number, line);
	warning(tempmsg);
    }
}


#if STDC
static char*
dupstr(const char *s)
#else
static char*
dupstr(s)
const char *s;
#endif
{			/* private replacement for non-standard strup() */
    char *p;

    p = (char*)malloc(strlen(s)+1);
    if (p == (char*)NULL)
	error("out of memory in dupstr()");
    return (strcpy(p,s));
}


static void
end_brace(VOID_ARG)
{
    out_yytext();
    adjust_brace_level(-1);
}


static void
end_bracket(VOID_ARG)
{
    out_yytext();
    adjust_bracket_level(-1);
}


static void
end_display_math(VOID_ARG)
{
    post_output_action = do_nothing;	/* prevent argument handling */
    remove_separating_blanks();
    adjust_environment_level(-1,yytext);
    if (default_end_display_math == (const char*)NULL)
	out_token_on_own_line(yytext);
    else
    {
	out_token_on_new_line(default_end_display_math);
	out_end_math_punctuation("%\n");
    }
    adjust_math_level(-1);
}


static void
end_environment(VOID_ARG)
{
    if (strncmp(yytext, "@end ", 5) == 0) /* TeXinfo @end can end either */
    {					/* an environment or a list */
	Style *style;

	style =  get_style_by_name(get_name(yytext));
	if ((style != (Style*)NULL) && (style->action == list))
	{
	    end_list();
	    return;
	}
    }

    post_output_action = do_nothing;	/* prevent argument handling */

    if (!Is_TeXinfo(yytext))		/* must protect "@end name" */
	remove_separating_blanks();

    /* Special cases: environment level is not altered by
       AmSTeX: \topmatter \endtopmatter and \document \enddocument
       LaTeX: \begin{document} \end{document}
    */
    if ((strcmp("\\end{document}", yytext) == 0) ||
	(strcmp("\\endtopmatter", yytext) == 0) ||
	(strcmp("\\enddocument", yytext) == 0))
	NOOP;
    else
	adjust_environment_level(-1,yytext);

    out_token_on_own_line(yytext);
    check_brace_level_is_zero("end of environment",0);
    check_bracket_level_is_zero("end of environment",0);
    check_parenthesis_level_is_zero("end of environment",0);
}


static void
end_list(VOID_ARG)
{
    post_output_action = do_nothing;	/* prevent argument handling */
    if (!Is_TeXinfo(yytext))		/* must protect "@end name" */
	remove_separating_blanks();
    adjust_environment_level(-2,yytext);
    out_token_on_own_line(yytext);

    check_brace_level_is_zero("end of list",0);
    check_bracket_level_is_zero("end of list",0);
    check_parenthesis_level_is_zero("end of list",0);
}


static void
end_math(VOID_ARG)
{
    post_output_action = do_nothing;	/* prevent argument handling */
    remove_separating_blanks();
    if (display_math_mode)
    {
	adjust_environment_level(-1,yytext);
	if (default_end_math == (const char*)NULL)
	    out_token_on_own_line(yytext);
	else
	{
	    out_token_on_new_line(default_end_math);
	    out_end_math_punctuation("%\n");
	}
    }
    else				/* inline math mode */
    {
	delete_spaces();
	out_blank();
	if (default_end_math == (const char*)NULL)
	    out_string(yytext);
	else
	{
	    out_string(default_end_math);
	    out_end_math_punctuation("");
	}
    }
    adjust_math_level(-1);
}


static void
end_parenthesis(VOID_ARG)
{
    out_yytext();
    adjust_parenthesis_level(-1);
}


static void
end_tabular(VOID_ARG)
{
    end_environment();
    in_tabular = 0;
}


#if STDC
static void
end_verbatim(const char *token)
#else
static void
end_verbatim(token)
const char *token;
#endif
{
    /* This function is NEVER called -- is it handled implicitly in
       copy_verbatim() */
    error("end_verbatim() function called in error");
    adjust_environment_level(-1,yytext);
    out_verbatim_string(token);
    out_newline();
}


static void
end_verbatim_output(VOID_ARG)
{
    in_verbatim = 0;
}


static void
environment(VOID_ARG)
{
    Style *style;
    StyleFunction the_action;

    remove_separating_blanks();
    style = get_style_by_name(get_name(yytext));

    /* With the exception of \egroup, environment terminators all */
    /* look like \end... or @end but list starters do not always look */
    /* like \begin....  If the action is set to environment, we */
    /* need to replace it by begin_environment or end_environment. */

    if ((style != (Style*)NULL) && (style->action != environment))
	the_action =  style->action;
    else if (strncmp("\\end",yytext,4) == 0)
	the_action = end_environment;
    else if (strncmp("@end ",yytext,5) == 0)
	the_action = end_environment;
    else if (strcmp("\\egroup",yytext) == 0)
	the_action = end_environment;
    else
	the_action = begin_environment;
    (*the_action)();
}


#if STDC
static void
error(const char *msg)
#else
static void
error(msg)
const char *msg;
#endif
{
    warning(msg);
    wrapup();
    exit (EXIT_FAILURE);
}


static void
free_hash_table(VOID_ARG)
{
    register int k;

    if (style_hash_table != (HASH_TABLE*)NULL)
    {
	for (k = 0; k < style_hash_table->hash_size; ++k)
	{
	    /* All hash_keys were dynamically allocated, so free them now */
	    if (style_hash_table->hash_table[k].hash_key != (const char*)NULL)
		free((VOIDP)style_hash_table->hash_table[k].hash_key);
	}
	hash_free(style_hash_table);
    }
}


static void
free_style_table(VOID_ARG)
{
    register int k;

    if (style_table.table != (Style*)NULL)
    {
	for (k = 0; k < style_table.size; ++k)
	    free((VOIDP)style_table.table[k].name);
	free((VOIDP)style_table.table);
    }
}


#if STDC
static char *
get_macro(const char *pattern)
#else
static char *
get_macro(pattern)
const char *pattern;
#endif
{		/* return a fresh copy of pattern with any args stripped */
    char *p;
    register char *q;

    p = q = dupstr(pattern);

    if (Is_Control_Sequence(q))
	++q;				/* move past escape character */
    while (Is_Control_Sequence_Character(*q))
	++q;
    *q = '\0';				/* strip trailing argument patterns */
    return (p);
}


#if STDC
VOIDP
get_memory(VOIDP ptr, size_t size)
#else
VOIDP
get_memory(ptr,size)
VOIDP ptr;
size_t size;
#endif
{
    /* allocate, or reallocate, memory, and return only if successful */

    if (ptr == (VOIDP)NULL)
	ptr = (VOIDP)malloc(size);
    else
	ptr = (VOIDP)realloc(ptr, size);
    if (ptr == (VOIDP )NULL)
    {
	error("out of memory in get_memory()");
	return ((VOIDP)NULL);		/* NOT REACHED */
    }
    else
	return (ptr);
}


#if STDC
static char *
get_name(const char *command)
#else
static char *
get_name(command)
const char *command;
#endif
{   /* Return environment name, or command name, in internal buffer which is */
    /* overwritten at each call. */

    static char *buffer = (char*)NULL;	/* dynamically-allocated buffer */
    int k;
    static size_t lenbuf = 0;		/* number of characters in buffer[] */
    size_t lenp;
    const char *p;
    char *q;

    static struct
    {
	const char *cmd;
	const char *name;
    } exception[] =		/* list of exceptional command->name mappings */
    {
	{ "\\bgroup",		"bgroup" },
	{ "\\egroup",		"bgroup" },
	{ "\\begingroup",	"begingroup" },
	{ "\\endgroup",		"begingroup" },
	{ (const char*)NULL,	(const char*)NULL }
    };

    /* command == \begin{name}, \end{name}, \name, \endname, \bgroup, */
    /* \egroup, \begingroup, \endgroup, $$, $(, $), $[, $], $$, */
    /* @name, @end name, or @name arg, possibly followed by punctuation. */

    for (p = (const char*)NULL, k = 0; exception[k].cmd != (const char*)NULL;
	 ++k)
    {					/* search exception list */
	if (strcmp(exception[k].cmd, command) == 0)
	{
	    p = exception[k].name;
	    break;
	}
    }
    if (p == (const char*)NULL)		/* command not in exception list */
    {
	if (!Is_TeXinfo(command))	/* might be "@itemize @minus{}" */
	    p = (const char *)strchr(command,'{'); /* '}' for brace balance */
	if (p == (const char *)NULL)	/* was not \begin{name} or \end{name} */
	{
	    p = command;
	    /* For AmSTeX-style \name and \endname, extract name */
	    if (Is_TeX(p) && Is_Control_Sequence_Character(p[1]))
		p = (strncmp(p+1,"end",3) == 0) ? (p + 4) : (p + 1);
	    else if (Is_TeXinfo(p)) /* reduce @end name and @name to @name */
	    {
		if (strncmp("@end ",p,5) == 0)
		    p += 5;
		for (NOOP; isspace(*p); ++p) /* skip over spaces */
		    NOOP;
		if ((p > command) && (p[-1] == ' '))
		    --p;		/* point to preceding space */
					/* so we can later deposit @ there */
	    }
	}
	else
	    p++;			/* point past open brace */
    }

    lenp = strlen(p);
    if (lenp >= lenbuf)			/* need to enlarge dynamic buffer */
    {
	lenbuf = lenp + 1;
	buffer = (char*)get_memory(buffer,lenbuf);
    }

    (void)strcpy(buffer,p);

    if (Is_TeXinfo(command))
    {
	if (buffer[0] == ' ')
	    buffer[0] = '@';		/* finish @end name -> @name reduction*/
	q = strchr(buffer,' ');
	if (q != (char*)NULL)		/* @name arg -> @name */
	    q[0] = '\0';
    }

    q = strchr(buffer,'}');
    if (q != (char*)NULL)		/* reduce possible {tabular}{|rrr|} */
	q[0] = '\0';			/* to tabular */
    else				/* remove trailing punctuation */
    {
	for (q = buffer;
	     *q && (Is_Control_Sequence_Character(*q) ||
		    (strchr("\\*$()[]",*q) != (char*)NULL));
	     ++q)
	    NOOP;
	*q = '\0';
    }

    return(buffer);
}


#if STDC
static Style *
get_style_by_name(const char *key)
#else
static Style *
get_style_by_name(key)
const char *key;
#endif
{
    HASH_ENTRY *h;

    h = hash_lookup(key, style_hash_table);

    return ((h == (HASH_ENTRY*)NULL) ? (Style*)NULL : (Style*)h->hash_data);
}


#if STDC
static Style *
get_style_by_style(const char *key)
#else
static Style *
get_style_by_style(key)
const char *key;
#endif
{
    register size_t k;

    for (k = 0; default_style[k].style != (const char*)NULL; ++k)
    {					/* match key against default styles */
	if (strcmp(default_style[k].style, key) == 0)
	    return (&default_style[k]);
    }

    return ((Style*)NULL);
}



#if STDC
static char*
get_token(char *name, const char *set, char *overwritten_char)
#else
static char*
get_token(name, set, overwritten_char)
char *name;
const char *set;
char *overwritten_char;
#endif
{		/* return next token made of chars in set[], or else NULL */
    register char *p;
    register char *q;

    for (p = name; *p && isspace(*p); ++p)
	NOOP;				/* move over leading whitespace */

    for (q = p; *q && (strchr(set,*q) != (char*)NULL); ++q)
	NOOP;				/* move over chars in set */

    *overwritten_char = *q;		/* save previous contents */
    *q = '\0';				/* terminate token */

    return ((p == q) ? (char*)NULL : p);
}


#if STDC
static HASH_INT
hash(const char *key, const HASH_TABLE *table)
#else
static HASH_INT
hash(key,table)
const char		*key;
const HASH_TABLE	*table;	/* hash_table created by hash_alloc() */
#endif
{
    HASH_INT		hash_code;

    if (table == (HASH_TABLE*)NULL)
	return (0);			/* NULL table: no index available */

    /* This hash function gives a reasonably even distribution when
       tested against the UNIX spelling dictionary, /usr/dict/words. */
    for (hash_code = 0; *key; ++key)
	hash_code = (hash_code << 1) ^ (0xff & *key);

    return (hash_code % (table->hash_size));
}


#if STDC
static HASH_TABLE*
hash_alloc(HASH_INT size)
#else
static HASH_TABLE*
hash_alloc(size)	/* allocate, initialize, and return hash table */
HASH_INT	size;	/* minimum required size of hash table */
#endif
{
    HASH_INT	k;
    HASH_TABLE	*table;

    size = (HASH_INT)next_prime((long)size);	/* make hash size a prime for */
						/* good hash performance */

    /* allocate hash table */
    table = (HASH_TABLE*)get_memory((VOIDP)NULL, size*sizeof(HASH_TABLE));

    /* initialize hash table */
    table->hash_index = (HASH_INT)0;
    table->hash_size = (HASH_INT)size;
    table->hash_table = (HASH_ENTRY*)get_memory((VOIDP)NULL,
						size*sizeof(HASH_ENTRY));

    /* initialize hash entries */
    for (k = 0; k < size; ++k)
    {
	table->hash_table[k].hash_key = (const char*)NULL;
	table->hash_table[k].hash_data = (VOIDP)NULL;
    }

    return (table);
}


#if STDC
static void
hash_free(HASH_TABLE *table)
#else
static void
hash_free(table)
HASH_TABLE	*table;
#endif
{
    if (table != (HASH_TABLE*)NULL)
    {
	free((VOIDP)table->hash_table);
	free((VOIDP)table);
    }
}


#if STDC
static HASH_ENTRY*
hash_install(const char *key, VOIDP data, HASH_TABLE *table)
#else
static HASH_ENTRY*
hash_install(key,data,table)
const char	*key;		/* key string tagging data */
VOIDP		data;		/* data to install */
HASH_TABLE	*table;		/* hash_table created by hash_alloc() */
#endif
{
    HASH_ENTRY	*h;

    h = hash_lookup(key,table);
    if (h != (HASH_ENTRY*)NULL)
    {
	/* NB: in texpretty, we are guaranteed that keys are dynamically
	   allocated, so if we are replacing an old entry, we should
	   free it to prevent a memory leak.  If this code is used
	   elsewhere, the free() call MUST be eliminated. */
	if (h->hash_key != (const char*)NULL)
	    free((VOIDP)h->hash_key);
	h->hash_key = key;
	h->hash_data = data;
    }
    return (h);
}


#if STDC
static HASH_ENTRY*
hash_lookup(const char *key, HASH_TABLE *table)
#else
static HASH_ENTRY*
hash_lookup(key,table)		/* return (maybe empty) entry for key, */
				/* or NULL if table full */
const char	*key;		/* key string to search for */
HASH_TABLE	*table;		/* hash_table created by hash_alloc() */
#endif
{
    HASH_ENTRY	*h = (HASH_ENTRY*)NULL;
    HASH_INT	n;
    HASH_INT	probes;

    if (table == (HASH_TABLE*)NULL)	/* NULL table has no entries */
	return ((HASH_ENTRY*)NULL);

    n = hash(key,table);		/* initial hash index */

    for (probes = 1; probes <= table->hash_size; ++probes)
    {
	if (table->hash_table[n].hash_key == (const char*)NULL)
	{
	    h = &table->hash_table[n];	/* not in table */
	    break;
	}
	else if (strcmp(table->hash_table[n].hash_key, key) == 0)
	{
	    h = &table->hash_table[n];	/* have entry in table */
	    break;
	}
	else
	{
	    n++;			/* linear probe to next entry */
	    if (n >= table->hash_size)
		n = (HASH_INT)0;
	}
    }
    return (h);
}


static int
indentation_size(VOID_ARG)
{
    int n;

    n = indent * environment_level;

    /* Ensure that a reasonable amount of space is left on each line, even */
    /* if the indentation level gets large.   Without this cap, out_char() */
    /* can produce an infinite loop(): out_char() -> out_leading_blanks() */
    /* out_blank() -> out_char(). */

    if (n > (max_width - 16))
	n = (max_width - 16);
    if (n < 0)
	n = 0;
    return (n);
}


static void
index_and_glossary(VOID_ARG)
{
    int c;
    int level;

    post_output_action = do_nothing;	/* prevent argument handling */
					/* because we do it specially here */
    comment_and_newline_before();	/* output macro name */

    for (;;)				/* discard space between macro */
    {					/* and argument */
	c = input_char();
	if (LEXEOF(c) || !isspace(c))
	    break;
    }

    if (c == '{')			/* '}' for brace balance */
    {					/* then found expected argument start */
	begin_verbatim_output();

	level = 0;
	do				/* copy braced argument */
	{
	    if (c == '{')
		level++;
	    else if (c == '}')
		level--;
	    out_verbatim_char(c);
	    c = input_char();
	}
	while ((level != 0) && !LEXEOF(c));

	switch (c)			/* examine lookahead character */
	{
	case '\n':			/* output a literal newline because */
					/* lex patterns would otherwise map */
					/* it to a space */
	    line_end();
	    if (last_char(0) == ' ')	/* did not find paragraph break */
	    {				/* so replace space by literal */
		trim_line(0);		/* newline */
		out_newline();
	    }
	    c = input_char();
	    break;

	case ' ':			/* another word follows, so we */
	    out_newline();		/* can safely insert a visible */
	    break;			/* line break */

	case '%':			/* user already dealt with followon */
	    break;

	default:			/* anything else following means we */
	    out_verbatim_char((int)'%'); /* can safely insert a hidden */
	    out_newline();		/* line break */
	    break;
	}

	end_verbatim_output();
    }
    if (!LEXEOF(c))
	unput_char(c);			/* restore lookahead */
}


static void
initialize(VOID_ARG)
{
    int k;
    unsigned const char *p;

#if defined(IBM_PC)
    /* On IBM PC DOS, use binary mode for stdin and stdout, so that Ctl-Z */
    /* does not mean end-of-file on input, and so that Ctl-M Ctl-J is not */
    /* written at end-of-line on output, causing doubled Ctl-M characters. */
    (void)setmode(fileno(stdin), O_BINARY);
    (void)setmode(fileno(stdout), O_BINARY);
#endif

    do_banner = 1;
    g_errors = 0;
    g_argk = 0;
    post_output_action = do_nothing;	/* prevent argument handling */

#define WRAP_BEFORE_CHARS " [<{("	/* ")}>]" for balance */

    for (p = (unsigned const char *)WRAP_BEFORE_CHARS; *p; ++p)
	wrap_before_chars[*p] = 1;

    for (k = 0; builtin_style[k] != (const char*)NULL; ++k)
    {
	char line[MAXBUF+2];		/* space for two trailing NULs */

	(void)strcpy(line,builtin_style[k]); /* get modifiable copy */
	do_style_line("internal-table", k, line);
    }
}


int
input_char(VOID_ARG)
{
    int c;

    c = Input();			/* this is the only place where the */

#if defined(IBM_PC)
    /* On PC DOS, map CR LF to LF on input, but allow bare CR too */
    if (c == '\r')
    {
	c = Input();
	if (c != '\n')			/* then have bare CR */
	{
	    unput_char(c);
	    c = '\r';
	}
    }
#endif

    if (c == '\n')			/* lex/flex input routine is called */
	line_number++;
    return (c);
}


#if STDC
int
last_char(int offset)
#else
int
last_char(offset)
int offset;
#endif
{	/* return the character offset characters from the last written */
	/* or EOF if not in buffer */
    int k;

    k = next_position - 1 + offset;

    if ((k < 0) || (next_position <= k))
	return (EOF);
    else
	return ((int)(0xff & line_buffer[k]));
}


static void
line_end(VOID_ARG)
{
    register int c;
    int original_line_number;

    /* One or more consecutive empty lines mark a paragraph boundary, */
    /* whereas a single newline is just replaced by a space. */

    original_line_number = line_number;

    for (c = '\n'; !LEXEOF(c) && isspace(c); c = input_char())
	NOOP;				/* gobble all following whitespace */
    if (!LEXEOF(c))
	unput_char(c);			/* put back non-space lookahead */
    if (line_number > original_line_number)
    {					/* have paragraph break */
	out_linebreak();		/* two or more newlines mark a */
	out_newline();			/* paragraph boundary */
	check_math_level_is_zero("paragraph end",0);
	check_brace_level_is_zero("end of paragraph",0);
	check_bracket_level_is_zero("end of paragraph",0);
	check_parenthesis_level_is_zero("end of paragraph",0);
    }
    else				/* have simple line break */
    {
	delete_spaces();		/* discard unneeded space and */
	out_blank();			/* replace line break by one space */
    }
}


static int
line_length(VOID_ARG)
{
    return (next_position);
}


static void
list(VOID_ARG)
{
    remove_separating_blanks();

    /* List terminators all look like "\end..." in LaTeX, plain TeX, */
    /* AmSTeX, and LAMsTeX, but list starters do not always look like */
    /* "\begin..." */

    if ((strncmp("\\end",yytext,4) == 0) || (strncmp("@end",yytext,4) == 0))
	end_list();
    else
	begin_list();
}


static void
list_item(VOID_ARG)
{
    remove_separating_blanks();

    adjust_list_level(-1,yytext);
    if (Is_TeXinfo(yytext))
    {			/* @item sometimes has argument on same line */
	newline_before();
	copy_to_end_of_line();
    }
    else
	out_token_on_own_line(yytext);
    adjust_list_level(1,yytext);
    check_brace_level_is_zero("start of list item",0);
    check_bracket_level_is_zero("start of list item",0);
    check_parenthesis_level_is_zero("start of list item",0);
}


#if STDC
static const char *
long_to_short_option(const char *option)
#else
static const char *
long_to_short_option(option)
const char *option;
#endif
{
    /* We support long GNU/POSIX style options by remapping them
       to the single-character options supported by texpretty 0.00 */
    static struct
    {
	const char *short_option;
	const char *long_option;
	size_t min_length;
    } option_table[] =
    {
      /* NB: This table assumes that "-" is always an acceptable
	 option prefix, independent of operating system! That is the
	 author's intent! */
	{ "-@",		"-texinfo",		3 },
	{ "-a",		"-author",		2 },
	{ "-c",		"-copyright",		2 },
	{ "-d",		"-displaymath-all",	2 },
	{ "-e",		"-emacs-mode",		2 },
	{ "-f",		"-filename",		2 },
	{ "-h",		"-help",		2 },
	{ "-i",		"-indent",		2 },
	{ "-l",		"-logfile",		2 },
	{ "-m",		"-math-conversions",	2 },
	{ "-n",		"-no-comment-banner",	2 },
	{ "-o",		"-outfile",		2 },
	{ "-q",		"-quick",		2 },
	{ "-s",		"-stylefile",		2 },
	{ "-t",		"-tabular-as-verbatim",	3 },
	{ "-v",		"-version",		2 },
	{ "-w",		"-width",		2 },
	{ (const char*)NULL, (const char*)NULL,	0 }
    };
    size_t k;
    size_t len;
    const char *org_option;

    org_option = option;

    if (OPTION_PREFIX(option[0]))
    {
	if (OPTION_PREFIX(option[1]))
	    ++option;			/* then have GNU/POSIX --option  */
	len = strlen(option);
	for (k = 0; option_table[k].long_option != (const char*)NULL; ++k)
	{
	    if (strnicmp(option,option_table[k].long_option,
			 (len > option_table[k].min_length) ?
			 len : option_table[k].min_length) == 0)
		return (option_table[k].short_option);
	}
    }

    return (org_option);
}


static void
make_hash_table(VOID_ARG)
{
    register int k;

    free_hash_table();			/* discard any previous hash table */

    style_hash_table = hash_alloc((HASH_INT)(10 + (style_table.size / 8) * 10));
					/* want table about 80% full */

    if (style_hash_table == (HASH_TABLE*)NULL)
	error("out of memory in make_hash_table()");

    for (k = 0; k < style_table.size; ++k)
    {
	(void)hash_install(get_macro(style_table.table[k].name),
			   (VOIDP)&style_table.table[k],
			   style_hash_table);
    }
}


static void
math(VOID_ARG)
{
    if (in_texinfo)
	out_yytext();
    else if (math_level > 0)
	end_math();
    else
	begin_math();
}


static void
newline_after(VOID_ARG)
{
    out_yytext();
    out_newline();
}


static void
newline_before(VOID_ARG)
{
    out_linebreak();
    out_yytext();
}


static long
#if STDC
next_prime(
long n
)
#else
next_prime(n)			/* return next prime at or after n */
long n;
#endif
{
    long prime;				/* tentative prime */
    long factor;			/* prime factor */
    int is_prime;			/* 'prime' is a prime number */

    n = (n < 0L) ? -n : n;		/* be safe -- force n positive */
    prime = 2L*(n/2L) + 1L;		/* next odd number */
    is_prime = (prime <= 3L);
    while (!is_prime)
    {
	factor = 5L;
	is_prime = (prime % 2L) && (prime % 3L);
	while (is_prime && (factor*factor <= prime))
	{
	    if ((prime % factor) == 0L)
		is_prime = 0;
	    else if ((prime % (factor + 2L)) == 0L)
		is_prime = 0;
	    else	/* factor+4 is divisible by 3 (every 3rd odd is) */
		factor += 6L;
	}
	if (!is_prime)
	    prime += 2L;
    }
    return (prime);
}


#if STDC
static void
normalize_space(char *s)
#else
static void
normalize_space(s)
char *s;
#endif
{					/* change all tabs in s[] to spaces */
    char *q;

    for (q = s; *s ; ++q, ++s)
    {
	if (isspace(*s))
	{
	    *q++ = ' ';			/* copy one space */
	    while (isspace(*s))		/* and skip the rest */
		++s;
	}
	*q = *s;			/* copy non-space */
	if (*s == '\0')			/* test for premature loop exit */
	    break;
    }
    *q = '\0';				/* terminate string */
}


static void
obey(VOID_ARG)
{
    register int c;

    /* When we meet \obeylines or \obeyspaces, we have no reliable */
    /* way to detect when their scope terminates, so we just */
    /* copy the rest of the file verbatim. */

    out_yytext();
    while ((c = input_char()), !LEXEOF(c))
	out_char(c);
}


static void
out_banner(VOID_ARG)
{
    char	timestr[27];
    long	the_timeval;

    in_comment = 1;

    out_string(COMMENT_PREFIX);
    out_string("-*-");
    out_string(emacs_mode);
    out_string("-*-");
    out_string(COMMENT_SUFFIX);
    out_newline();

    if (filename && *filename)
    {
	out_string(COMMENT_PREFIX);
	out_string(filename);
	out_string(COMMENT_SUFFIX);
	out_newline();
    }

    out_string(COMMENT_PREFIX);
    out_string("Prettyprinted by ");
    out_string(PROGRAM);
    out_string(" lex version ");
    out_string(VERSION);
    out_string(" ");
    out_string(DATE);
    out_string(COMMENT_SUFFIX);
    out_newline();

    the_timeval = time((time_t *)NULL);
    (void)strcpy(timestr,ctime((const time_t*)(&the_timeval)));
    timestr[24] = '\0';			/* kill \n from ctime() */
    out_string(COMMENT_PREFIX);
    out_string("on ");
    out_string(timestr);
    out_string(COMMENT_SUFFIX);
    out_newline();

#if defined(HAVE_PWD_H)
    {
	const char *login_name;
	const char *personal_name;
	struct passwd *pwd;

#if defined(HAVE_UNAME)
	struct utsname uts;
#else
	char hostname[256+1];
#endif

	login_name = (const char*)getlogin();
	personal_name = "";
	if (login_name == (const char*)NULL)
	    pwd = (struct passwd *)getpwuid(getuid());
	else
	    pwd = (struct passwd *)getpwnam(login_name);
	if (pwd != (struct passwd *)NULL)
	{
	    login_name = (const char*)pwd->pw_name;
	    personal_name = (const char*)pwd->pw_gecos;
	}
	else
	{
	    login_name = "";
	    personal_name = "";
	}

#if defined(HAVE_UNAME)
	if (uname(&uts) < 0)
	    (void)strcpy(uts.nodename,"");
#else
	if (gethostname(hostname, sizeof(hostname)) < 0)
	    (void)strcpy(hostname, "");
#endif

	out_string(COMMENT_PREFIX);
	out_string("for ");
	out_string(personal_name);
	out_string(" (");
	out_string(login_name);
	out_string("@");

#if defined(HAVE_UNAME)
	out_string(uts.nodename);
#else
	out_string(hostname);
#endif

	out_string(")");
	out_string(COMMENT_SUFFIX);
	out_newline();
    }
#endif /* defined(HAVE_PWD_H) */

    out_newline();			/* supply one blank line after banner */
    in_comment = 0;
}


static void
out_blank(VOID_ARG)
{			/* output a blank, except at beginning of line */
    if (Is_NonEmpty_Line()) /* then not at start of line, */
	out_char((int)' ');
}


static void
out_buffer(VOID_ARG)
{
    if (next_position > 0)
    {
	line_buffer[next_position] = '\0';
	if (fputs(line_buffer, stdout) == EOF)
	    error("output failure: perhaps disk full?");
	next_position = 0;
	line_buffer[next_position] = '\0'; /* for DEBUGGING only */
    }
}


#if STDC
void
out_char(int c)
#else
void
out_char(c)
int c;
#endif
{		/* ALL output goes through this routine, but it is buffered */
		/* in memory until a newline is written. */
    int wrapping;

    if (in_comment || in_verbatim)
	wrapping = 0;
    else if ((next_position >= max_width) && (c == (int)' '))
    {					/* break long line here */
	c = '\n';
	wrapping = 1;
    }
    else if (c == '\n')
	wrapping = 1;
    else
	wrapping = 0;

    if (c == '\n')
    {
	if (in_verbatim)
	    NOOP;			/* trailing space must be preserved */
	else
	    trim_line(0);		/* trim trailing space at end-of-line */

	if (in_comment || in_verbatim)
	    NOOP;
	else
	{
	    /* open brace at end of line is undesirable, although */
	    /* rarely, a following space may be intended.  */
	    if ((next_position > 0) && (line_buffer[next_position-1] == '{'))
		line_buffer[next_position++] = '%';
	}

#if defined(IBM_PC)
	/***************************************************************
	On the IBM PC, we handle input and output in binary mode, and
	input \r (Ctl-M) gets mapped to blank by a lex pattern.  Thus,
	on output, we need to map \n (Ctl-J) back to \r\n to match DOS
	conventions.
	***************************************************************/
	line_buffer[next_position++] = '\r';
#endif

    }

    line_buffer[next_position++] = c;

    line_buffer[next_position] = '\0';	/* for DEBUGGING only */

    /* next_position must not advance beyond MAXBUF-4, because there must */
    /* be room to deposit four characters ('%', '\r', '\n', '\0') on the */
    /* next call to this function */
    if ((c == '\n') || (next_position >= (MAXBUF - 4)))
    {					/* buffer full, or end of line */
	out_buffer();
	if (in_comment || in_verbatim)
	    NOOP;
	else if (wrapping)
	    out_indentation();
    }
}


static void
out_comment(VOID_ARG)
{			/* output a comment, INCLUDING the following newline, */
			/* AND guaranteeing preservation of the newline */
    register int c;
    register int c_last;
    register int k;

    for (c_last = 0, k = 0; yytext[k]; ++k)
    {
	if ((c_last != (int)'\\') && (c_last != (int)'@'))
	    check_delimiter_level((int)yytext[k]);
	c_last = (int)yytext[k];
    }
    in_comment = 1;
    post_output_action = do_nothing;	/* prevent argument handling */
    out_yytext();
    for (c_last = 0, c = input_char(); !LEXEOF(c) && (c != '\n');
	 c = input_char())
    {
	if ((c_last != (int)'\\') && (c_last != (int)'@'))
	    check_delimiter_level(c);
	out_char(c);			/* copy comment to end-of-line */
	c_last = c;
    }
    if (c == '\n')
    {
	line_end();			/* this might just insert a space */
	out_linebreak();		/* so trim it and then ensure */
					/* that we get a true newline */
    }
    in_comment = 0;
}


static void
out_control_sequence(VOID_ARG)
{
    Style *style;

    style = get_style_by_name(yytext);
    if (style == (Style*)NULL)
    {
	if (in_texinfo && Is_TeXinfo(yytext))
	{

#undef MSG_FMT
#define MSG_FMT "unrecognized TeXinfo command %.39s"

	    char tempmsg[sizeof(MSG_FMT) + 39 + 1];

	    SPRINTF(tempmsg, MSG_FMT, yytext);
	    warning(tempmsg);
	}
	out_yytext();
    }
    else
    {
	post_output_action = do_macro_arguments;
	(*style->action)();
    }
}


#if STDC
static void
out_end_math_punctuation(const char *prefix)
#else
static void
out_end_math_punctuation(prefix)
const char *prefix;
#endif
{
    const char *punctuation;

    if (strncmp(yytext,"\\]",2) == 0)
	punctuation = &yytext[2];
    else if (strncmp(yytext,"\\)",2) == 0)
	punctuation = &yytext[2];
    else if (strncmp(yytext,"$$",2) == 0)
	punctuation = &yytext[2];
    else if (strncmp(yytext,"$",1) == 0)
	punctuation = &yytext[1];
    else if (strncmp(yytext,"\\end",4) == 0)
    {
	punctuation = strchr(yytext,'}');
	if (punctuation == (const char*)NULL)
	    punctuation = "";		/* should never happen */
	else
	    punctuation++;
    }
    else
	punctuation = "";
    if (*punctuation)
    {
	if ((prefix != (const char*)NULL) && *prefix)
	    out_string(prefix);
	out_string(punctuation);
    }
}


void
out_indentation(VOID_ARG)
{
    trim_line(0);
    out_leading_blanks(indentation_size());
}


#if STDC
static void
out_leading_blanks(int nblanks)
#else
static void
out_leading_blanks(nblanks)
int nblanks;
#endif
{

#if defined(DEBUG)
    SPRINTF(prefix, "%02d:", environment_level); /* for DEBUGGING */
#endif

    for ( ; nblanks > 0; --nblanks)
	out_char((int)' ');		/* NB: must NOT be out_blank() */
					/* because that function suppresses */
					/* blanks at start of line */
}


void
out_linebreak(VOID_ARG)
{				/* output a newline, except at start of line */
    trim_line(0);
    if (line_length() > 0)
	out_newline();
}


#if STDC
static void
out_linenumbers(const int *array, const char **name, int n)
#else
static void
out_linenumbers(array,name,n)
const int *array;
const char **name;
int n;
#endif
{
    register int count;
    register int k;

    for (count = 0, k = 1; k <= n; ++k)
    {
	if ((count == 0) && (k > 1))
	    PUTSTDERR("\t\t");
	FPRINTF(stderr," %d", array[k]);
	if ((name != (const char**)NULL) && (name[k] != (const char*)NULL))
	    FPRINTF(stderr," {%s}", name[k]);
	count++;
	if ((k < n) && ((k % 5) == 0))
	{
	    PUTSTDERR("\n");
	    count = 0;
	}
    }
    if (count > 0)
	PUTSTDERR("\n");
}


static void
out_newline(VOID_ARG)
{
    out_char((int)'\n');
}


#if STDC
void
out_string(const char *s)
#else
void
out_string(s)
const char *s;
#endif
{	/* output a string, line wrapping at WRAP_BEFORE_CHARS if necessary */
    const char *original_s;
    const char *p;
    char prev_s;
    PostOutputAction action;

    if (line_length() < indentation_size()) /* first token on this line */
	out_indentation();
    if (line_length() == indentation_size())
    {
	for (p = s; *p == ' '; ++p)	/* does s match /^[ ]+$/ ? */
	    NOOP;
	if (*p == '\0')		/* yes, it does */
	    return;		/* discard blank string at beginning of line */
    }
    for (original_s = s, prev_s = '\0' ; *s ; prev_s = *s, ++s)
    {
	if (in_comment || in_verbatim)	/* no line wrapping permitted */
	    NOOP;
	else if (prev_s == '\\')
	    NOOP;
	else if (prev_s == '@')
	    NOOP;
	else if (prev_s == '@')
	    NOOP;
#if 0
	else if (strchr(WRAP_BEFORE_CHARS,*s) != (char*)NULL)
#else
	else if (wrap_before_chars[(unsigned char)*s])
#endif
	{			/* then possible place to wrap the line */
	    if (((line_length() + 1 +
		  (int)strcspn(s+1,WRAP_BEFORE_CHARS)) > max_width) &&
		Is_NonEmpty_Line())
	    {			/* then line must be wrapped */
		out_newline();
		if (s > original_s)
		    adjust_environment_level(1,""); /* indent wrapped lines one more level */
		out_indentation();
		if (s > original_s)
		    adjust_environment_level(-1,"");
		if (*s == ' ')
		{
		    for (++s; *s && isspace(*s); ++s)
			NOOP;		/* skip over repeated spaces */
		    s--;		/* and point back to last space */
		    continue;
		}
	    }
	}
	out_char(*s);
    }
    if (post_output_action != do_nothing)
    {
	Style *style;

	style = get_style_by_name(original_s);
	if (style != (Style*)NULL)
	{
	    action = post_output_action;	/* to prevent recursive */
	    post_output_action = do_nothing;	/* application of the action */
	    (*action)(style->name);
	}
    }
}


#if STDC
static void
out_token_on_new_line(const char *token)
#else
static void
out_token_on_new_line(token)
const char *token;
#endif
{
    if (!Is_TeXinfo(token))
	remove_separating_blanks();
    delete_spaces();
    if (Is_NonEmpty_Line())		/* then put this token on new line */
	out_newline();
    out_string(token);
}


#if STDC
static void
out_token_on_own_line(const char *token)
#else
static void
out_token_on_own_line(token)
const char *token;
#endif
{
    out_token_on_new_line(token);
    out_newline();
}


#if STDC
static void
out_verbatim_string(const char *s)
#else
static void
out_verbatim_string(s)
const char *s;
#endif
{
    for (; *s; ++s)
	out_verbatim_char((int)*s);
}


static void
out_yytext(VOID_ARG)
{
    if ((last_char(0) == ' ') &&
	((line_length() + (int)strlen(yytext)) > max_width))
	out_newline();			 /* insufficient space for this token */
    out_string(yytext);
}


static void
remove_separating_blanks(VOID_ARG)
{		/* remove blanks in yytext[],  provided that they occur at */
		/* brace/paren/bracket level zero */
    register char *p;
    register char *q;
    int the_brace_level;
    int the_bracket_level;
    int the_parenthesis_level;

    the_brace_level = 0;
    the_bracket_level = 0;
    the_parenthesis_level = 0;

    for (q = p = &yytext[0]; *p ; ++p)
    {
	switch (*p)
	{
	case '{':
	    the_brace_level++;
	    break;

	case '}':
	    the_brace_level--;
	    break;

	case '[':
	    the_bracket_level++;
	    break;

	case ']':
	    the_bracket_level--;
	    break;

	case '(':
	    the_parenthesis_level++;
	    break;

	case ')':
	    the_parenthesis_level--;
	    break;

	default:
	    break;
	}
	if (isspace(*p) &&
	    (the_brace_level == 0) &&
	    (the_bracket_level == 0) &&
	    (the_parenthesis_level == 0))
	    NOOP;			/* drop this space */
	else
	    *q++ = *p;			/* copy character */
    }
    *q = '\0';				/* terminate result */
}


static void
sectional_division(VOID_ARG)
{
    check_brace_level_is_zero("start of sectional division",1);
    check_bracket_level_is_zero("start of sectional division",1);
    check_parenthesis_level_is_zero("start of sectional division",1);
    check_math_level_is_zero("start of sectional division",1);
    check_environment_level_is_zero("start of sectional division",1);
    newline_before();
}


static void
sectional_division_line(VOID_ARG)
{
    check_brace_level_is_zero("start of sectional division",1);
    check_bracket_level_is_zero("start of sectional division",1);
    check_parenthesis_level_is_zero("start of sectional division",1);
    check_math_level_is_zero("start of sectional division",1);
    check_environment_level_is_zero("start of sectional division",1);
    newline_before();
    copy_to_end_of_line();
}


static void
standalone(VOID_ARG)
{
    out_token_on_own_line(yytext);
}


static void
tabular(VOID_ARG)
{
    if (strncmp("\\begin",yytext,6) == 0)
	begin_tabular();
    else if (strncmp("\\end",yytext,4) == 0)
	end_tabular();
    else
    {
	warning("unrecognized tabular environment"); /* should never happen */
	post_output_action = do_nothing;	/* prevent argument handling */
	out_yytext();
    }
}


#if STDC
static void
trim_line(int n)
#else
static void
trim_line(n)
int n;
#endif
{
    for (; (next_position > 0) && (next_position > n) &&
	((line_buffer[next_position-1] == ' ') ||
#if defined(IBM_PC)
	(line_buffer[next_position-1] == '\r') ||
#endif
	(line_buffer[next_position-1] == '\t'));
	 --next_position)
	NOOP;
}


#if STDC
static void
unput_char(int c)
#else
static void
unput_char(c)
int c;
#endif
{
    if (c == '\n')
	line_number--;
    if (!LEXEOF(c))
	unput(c);			/* this is the only place where the */
					/* lex/flex unput() routine is called */
}


static void
usage(VOID_ARG)
{
    static int first = 1;

    if (first)
    {
	PUTSTDERR(g_argv[0]);
	PUTSTDERR(" [-?] [-@] [-a] [-c] [-d] [-e mode] [-f filename] [-h]\n");
	PUTSTDERR("\t\t[-i nnn] [-l filename] [-m nnn] [-n] [-o filename]\n");
	PUTSTDERR("\t\t[-q] [-s filename] [-t] [-v] [-w width]\n");
	PUTSTDERR("\t\t[file(s)]\n");
	PUTSTDERR("or\n");
	PUTSTDERR(g_argv[0]);
	PUTSTDERR(" [--?] [--texinfo] [--author] [--copyright] [--displaymath-all]\n");
	PUTSTDERR("\t\t[--emacs-mode mode] [--filename filename] [--help]\n");
	PUTSTDERR("\t\t[--indent nnn] [--logfile filename] [--math-conversions nnn]\n");
	PUTSTDERR("\t\t[--no-comment-banner] [--outfile filename] [--quick]\n");
	PUTSTDERR("\t\t[--stylefile filename] [--tabular-as-verbatim] [--version]\n");
	PUTSTDERR("\t\t[--width width]\n");
	PUTSTDERR("\t\t[file(s)]\n");
	PUTSTDERR("where the options are:\n");
	PUTSTDERR("\t[-?]          display usage summary\n");
	PUTSTDERR("\t[-@]          set TeXinfo mode\n");
	PUTSTDERR("\t[-a]          show author credit\n");
	PUTSTDERR("\t[-c]          show copyright information\n");
	PUTSTDERR("\t[-d]          output inline math like display math\n");
	PUTSTDERR("\t[-e mode]     set Emacs editor mode name\n");
	PUTSTDERR("\t[-f filename] set output filename in comment banner\n");
	PUTSTDERR("\t[-h]          display usage summary\n");
	PUTSTDERR("\t[-i nnn]      set indentation to nnn spaces per level\n");
	PUTSTDERR("\t[-l filename] set error log filename\n");
	PUTSTDERR("\t[-m nnn]      set math conversions\n");
	PUTSTDERR("\t              nnn = sum of:\n");
	PUTSTDERR("\t               1\tinline math  -> $...$\n");
	PUTSTDERR("\t               2\tinline math  -> \\(...\\)\n");
	PUTSTDERR("\t               4\tinline math  -> \\begin{math}...\\end{math}\n");
	PUTSTDERR("\t               8\tdisplay math -> $$...$$\n");
	PUTSTDERR("\t              16\tdisplay math -> \\[...\\]\n");
	PUTSTDERR("\t              32\tdisplay math -> \\begin{displaymath}...\\end{displaymath}\n");
	PUTSTDERR("\t[-n]          no comment banner\n");
	PUTSTDERR("\t[-o filename] set output filename\n");
	PUTSTDERR("\t[-q]          no default style files\n");
	PUTSTDERR("\t[-s filename] set style filename\n");
	PUTSTDERR("\t[-t]          copy tabular environments verbatim\n");
	PUTSTDERR("\t[-v]          show version number\n");
	PUTSTDERR("\t[-w nnn]      set output line width to nnn\n");
    }
    first = 0;
}


static void
verb(VOID_ARG)
{
    register int c;
    register int delimiter;
    int doubled_delimiter_okay;
    int newline_okay_in_verbatim;
    register int v;

#if defined(IBM_PC)
    static char verbatim_buffer[MAXBUF]; /* conserve stack space */
#else
    char verbatim_buffer[MAXBUF];
#endif

    post_output_action = do_nothing;	/* prevent argument handling */
    v = (int)strlen(yytext);
    if (v >= sizeof(verbatim_buffer))
	error("verbatim_buffer[] too small"); /* should never happen */

    (void)strcpy(verbatim_buffer,yytext);

    /* LAmSTeX \Lit*...* preserves newlines exactly */
    newline_okay_in_verbatim = (strncmp(yytext,"\\Lit",4) == 0);

    /* ETeX \verbatim|...|endverbatim allows doubled delimiter and newlines */
    doubled_delimiter_okay = (strncmp(yytext,"\\verbatim",9) == 0);
    if (doubled_delimiter_okay)
	newline_okay_in_verbatim = 1;

    c = input_char();
    if (LEXEOF(c))
	NOOP;
    else if (Is_Control_Sequence_Character(c) || isspace(c) ||
	     ((strncmp(&yytext[1],"verb",4) == 0) && (c == '*')))
	unput_char(c);			/* false match */
    else
    {
	verbatim_buffer[v++] = (char)c;
	delimiter = c;
	do
	{
	    c = input_char();
	    if (v >= (sizeof(verbatim_buffer) - 2)) /* need room for two more */
	    {			/* in case of doubled_delimiter_okay */
		if (((line_length() + v) > max_width) && !in_verbatim)
		    out_newline();
		begin_verbatim_output();
		verbatim_buffer[v] = '\0';
		out_verbatim_string(verbatim_buffer);
		/* NB: no end_verbatim_output() here, because we need */
		/* to prevent any further out_newline() calls */
		v = 0;
	    }
	    if ((c == '\n') && !newline_okay_in_verbatim)
	    {
		warning("newline in verbatim string changed to space");
		c = ' ';
	    }
	    verbatim_buffer[v++] = (char)c;
	    if ((c == delimiter) && doubled_delimiter_okay)
	    {
		c = input_char();
		if (c == delimiter)	/* then have doubled delimiter */
		{
		    verbatim_buffer[v++] = (char)c;
		    c = 256;		/* to prevent loop termination */
		}
		else			/* restore lookahead */
		{
		    unput_char(c);
		    c = delimiter;
		}
	    }
	} while ((c != delimiter) && !LEXEOF(c));
    }

    if (((line_length() + v) > max_width) && !in_verbatim)
	out_newline();
    begin_verbatim_output();
    verbatim_buffer[v] = '\0';
    out_verbatim_string(verbatim_buffer);
    end_verbatim_output();
}


static void
verbatim(VOID_ARG)
{
    post_output_action = do_nothing;	/* prevent argument handling */
    remove_separating_blanks();
    if ((yytext[1] == 'b') || Is_TeXinfo(yytext))
    {				/* expect \begin... or @name (TeXinfo) */
	begin_verbatim();
	copy_verbatim();  /* handles \end{verbatim} and \end{verbatim*} */
    }
    else
	end_verbatim((const char*)&yytext[0]); /* never actually called */
}


static void
version(VOID_ARG)
{
    FPRINTF(stderr, "TeX prettyprinter %s version %s\n",
	    PROGRAM, VERSION);
    FPRINTF(stderr, "Compiled on %s %s by %s@%s\n",
	    __DATE__, __TIME__, USER, HOSTNAME);
}


#if STDC
static void
warning(const char *msg)
#else
static void
warning(msg)
const char *msg;
#endif
{
    FPRINTF(stderr,"%s:%d:%s\n",
	    (filename == (const char*)NULL) ? "" : filename, line_number, msg);
    g_errors++;
}


static void
wrapup(VOID_ARG)
{
    out_linebreak();			/* force out last line */
    out_buffer();

    check_math_level_is_zero("end of file",0);
    check_brace_level_is_zero("end of file",0);
    check_bracket_level_is_zero("end of file",0);
    check_parenthesis_level_is_zero("end of file",0);
    check_environment_level_is_zero("end of file",0);

    free_hash_table();			/* NB: order here is CRITICAL: free */
    free_style_table();			/* hash table BEFORE style table! */
}


#if STDC
int
main(int argc, char *argv[])
#else
int
main(argc, argv)
int argc;
char *argv[];
#endif
{
    g_argc = argc;
    g_argv = &argv[0];

    initialize();

    do_initialization_files();

    make_hash_table();

    (void)reopen();

    if (do_banner)
	out_banner();

    while (yylex())
	NOOP;

    wrapup();				/* to force out last buffer */

    return ((g_errors > 0) ? EXIT_FAILURE : EXIT_SUCCESS);
}


static void
do_command_line_argument(VOID_ARG)
{
    int n;
    const char *option;

    option = long_to_short_option(g_argv[g_argk]);

    if (option[2] == '\0')		/* then have valid -x option */
    {
	switch (option[1])
	{
	case '@':
	    in_texinfo = 1;
	    break;

	case 'a':			/* show author */
	case 'A':
	    author();
	    exit(EXIT_SUCCESS);
	    break;

	case 'c':			/* show copyright */
	case 'C':
	    copyright();
	    exit(EXIT_SUCCESS);
	    break;

	case 'd':			/* inline math like display math */
	case 'D':
	    display_math_mode = 1;
	    break;

	case 'e':			/* set emacs modename */
	case 'E':
	    if (g_argv[++g_argk] != (char*)NULL)
		emacs_mode = (const char*)g_argv[g_argk];
	    else
	    {
		warning("-e option missing emacs-mode value");
		usage();
		exit(EXIT_FAILURE);
	    }
	    break;

	case 'f':		/* set output filename in comment banner */
	case 'F':
	    filename = (const char*)g_argv[++g_argk];
	    if (filename == (const char *)NULL)
	    {
		warning("-f option missing comment-banner filename");
		usage();
		exit(EXIT_FAILURE);
	    }
	    break;

	case 'h':			/* show help message */
	case 'H':
	case '?':
	    usage();
	    exit(EXIT_SUCCESS);
	    break;

	case 'i':			/* set indentation size */
	case 'I':
	    if ((g_argv[++g_argk] != (char*)NULL) && isdigit(g_argv[g_argk][0]))
	    {
		n = (int)atoi(g_argv[g_argk]);
		indent = MAX(0,MIN(n,16));	/* require reasonable range */
	    }
	    else
	    {
		warning("-i option missing indentation value");
		usage();
		exit(EXIT_FAILURE);
	    }
	    break;

	case 'l':			/* set error log filename */
	case 'L':
	    if (g_argv[++g_argk] != (char*)NULL)
	    {
		if (freopen(g_argv[g_argk], WMODE, stderr) == (FILE*)NULL)
		    exit(EXIT_FAILURE);	/* no stderr left for warning() */
	    }
	    else
	    {
		warning("-l option missing log filename");
		usage();
		exit(EXIT_FAILURE);
	    }
	    break;

	case 'm':			/* math mode conversions */
	case 'M':
	    if ((g_argv[++g_argk] != (char*)NULL) && isdigit(g_argv[g_argk][0]))
	    {
		math_conversions = (int)atoi(g_argv[g_argk]);
		if (math_conversions & 1)
		{
		    default_begin_math = "$";
		    default_end_math   = "$";
		}
		if (math_conversions & 2)
		{
		    default_begin_math = "\\(";
		    default_end_math   = "\\)";
		}
		if (math_conversions & 4)
		{
		    default_begin_math = "\\begin{math}";
		    default_end_math   = "\\end{math}";
		}
		if (math_conversions & 8)
		{
		    default_begin_display_math = "$$";
		    default_end_display_math   = "$$";
		}
		if (math_conversions & 16)
		{
		    default_begin_display_math = "\\[";
		    default_end_display_math   = "\\]";
		}
		if (math_conversions & 32)
		{
		    default_begin_display_math = "\\begin{displaymath}";
		    default_end_display_math   = "\\end{displaymath}";
		}
	    }
	    else
	    {
		warning("-m option missing math conversions value");
		usage();
		exit(EXIT_FAILURE);
	    }
	    break;

	case 'n':			/* no comment banner */
	case 'N':
	    do_banner = 0;
	    break;

	case 'o':			/* alternate output file */
	case 'O':
	    if (g_argv[++g_argk] != (char*)NULL)
	    {
		if (freopen(g_argv[g_argk], WMODE, stdout) == (FILE*)NULL)
		{
		    warning("-o outfile reopen failed");
		    exit(EXIT_FAILURE);
		}
	    }
	    else
	    {
		warning("-o option missing output filename");
		usage();
		exit(EXIT_FAILURE);
	    }
	    break;

	case 'q':			/* no default style files */
	case 'Q':
	    break; /* already handled in main() -> do_initialization_files() */

	case 's':			/* set style filename */
	case 'S':
	    if (g_argv[++g_argk] != (char*)NULL)
	    {
		do_style_file(g_argv[g_argk],1);
		make_hash_table();
	    }
	    else
	    {
		warning("-s option missing style filename");
		usage();
		exit(EXIT_FAILURE);
	    }
	    break;

	case 't':			/* copy tabular environments verbatim */
	case 'T':
	    tabular_like_verbatim = 1;
	    break;

	case 'v':			/* show version */
	case 'V':
	    version();
	    exit(EXIT_SUCCESS);
	    break;

	case 'w':			/* set linewidth */
	case 'W':
	    if ((g_argv[++g_argk] != (char*)NULL) && isdigit(g_argv[g_argk][0]))
	    {
		n = (int)atoi(g_argv[g_argk]);
		max_width = MAX(16,MIN(n,MAXBUF)); /* require reasonable range */
	    }
	    else
	    {
		warning("-w option missing width value");
		usage();
		exit(EXIT_FAILURE);
	    }
	    break;

	default:
	    bad_option(g_argv[g_argk]);
	    break;
	}
    }
    else
	bad_option(g_argv[g_argk]);

    if (in_texinfo)
    {
	/* Turn off options that conflict with TeXinfo; sadly, */
	/* indentation is among them. */
	display_math_mode = 0;
	emacs_mode = "texinfo";
	indent = 0;
	math_conversions = 0;
	default_begin_display_math = (const char*)NULL;
	default_end_display_math = (const char*)NULL;
	default_begin_math = (const char*)NULL;
	default_end_math = (const char*)NULL;
    }
}


static void
do_initialization_files(VOID_ARG)
{
    /* Careful: some systems (e.g. HP-UX 9.0) have ridiculously small limits,
       so we insist that it be at least MAXBUF */
#if defined(FILENAME_MAX) && ((FILENAME_MAX + 0) > MAXBUF)
    char the_filename[FILENAME_MAX];
#else
    char the_filename[MAXBUF];
#endif

    int k;
    char *p;

    /* Prescan the options to see if -q (suppress default style files)
       has been specified. */

    for (k = 0; k < g_argc; ++k)
    {
	if (OPTION_PREFIX(g_argv[k][0]) &&
	    ((g_argv[k][1] == 'q') || (g_argv[k][1] == 'Q')) &&
	    (g_argv[k][2] == '\0'))
	    return;			/* yes, found -q option */
    }

    /* Try to read two initialization files: one in the home directory,
       and the other in the current directory. */

#if defined(IBM_PC) || defined(VMS)
    p = HOME_DIRECTORY;
#else
    p = getenv(HOME_DIRECTORY);
#endif

    if (p != (char*)NULL)
    {
	(void)strncpy(the_filename, p,
		      sizeof(the_filename) - sizeof(DEFAULT_STYLE_FILE) -
		      sizeof(DIRECTORY_SEPARATOR));
	the_filename[sizeof(the_filename)- sizeof(DEFAULT_STYLE_FILE) -
		 sizeof(DIRECTORY_SEPARATOR) - 1] = '\0';
	(void)strcat(the_filename,DIRECTORY_SEPARATOR);
	(void)strcat(the_filename,DEFAULT_STYLE_FILE);
	do_style_file(the_filename, 0);
    }

    (void)strncpy(the_filename, DEFAULT_STYLE_FILE, sizeof(DEFAULT_STYLE_FILE));
    the_filename[sizeof(the_filename)-1] = '\0';
    do_style_file(the_filename, 0);
}


static int
reopen(VOID_ARG)
{
    if (g_argk < (g_argc - 1))
    {
	++g_argk;
	if (OPTION_PREFIX(g_argv[g_argk][0]))
	{
	    do_command_line_argument();
	    return (reopen());
	}
	else if (freopen(g_argv[g_argk], RMODE, stdin) != (FILE*)NULL)
	{
	    filename = g_argv[g_argk];
	    line_number = 1;
	    return (0);			/* have file to process */
	}
	else
	{

#undef MSG_FMT
#define MSG_FMT "open failure on input file %.255s"

	    char tempmsg[sizeof(MSG_FMT) + 255 + 1];

	    SPRINTF(tempmsg, MSG_FMT, g_argv[g_argk]);
	    filename = (const char*)NULL;
	    warning(tempmsg);
	    return (reopen());
	}
    }
    else
	return (1);			/* all done */
}


#ifdef __cplusplus
extern "C"
#endif
int
yywrap(VOID_ARG)
{
    return (reopen());
}


#if defined(DEBUG_MALLOC)
#undef free
#if STDC
static void
debug_free(VOIDP ptr)
#else
static void
debug_free(ptr)
VOIDP ptr;
#endif
{
    VOIDP p;

    FPRINTF(stderr,"DEBUG: free(%p)\n", ptr);

    free(ptr);
}


#undef malloc
#if STDC
static VOIDP
debug_malloc(size_t size)
#else
static VOIDP
debug_malloc(size)
size_t size;
#endif
{
    VOIDP p;

    p = malloc(size);

    FPRINTF(stderr,"DEBUG: malloc(%ld) -> %p\n", (long)size, p);

    return (p);
}


#undef realloc
#if STDC
static VOIDP
debug_realloc(VOIDP ptr, size_t size)
#else
static VOIDP
debug_realloc(ptr, size)
VOIDP ptr;
size_t size;
#endif
{
    VOIDP p;

    p = realloc(ptr,size);

    FPRINTF(stderr,"DEBUG: realloc(%p,%ld) -> %p\n", ptr, (long)size, p);

    return (p);
}
#endif /* defined(DEBUG_MALLOC) */
